"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import tensorflow_metadata.proto.v0.path_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class DatasetFeatureStatisticsList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    DATASETS_FIELD_NUMBER: builtins.int

    @property
    def datasets(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___DatasetFeatureStatistics]: ...

    def __init__(self,
        *,
        datasets : typing.Optional[typing.Iterable[global___DatasetFeatureStatistics]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"datasets",b"datasets"]) -> None: ...
global___DatasetFeatureStatisticsList = DatasetFeatureStatisticsList

class DatasetFeatureStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    NUM_EXAMPLES_FIELD_NUMBER: builtins.int
    WEIGHTED_NUM_EXAMPLES_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    CROSS_FEATURES_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    num_examples: builtins.int = ...
    weighted_num_examples: builtins.float = ...

    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FeatureNameStatistics]: ...

    @property
    def cross_features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___CrossFeatureStatistics]: ...

    def __init__(self,
        *,
        name : typing.Text = ...,
        num_examples : builtins.int = ...,
        weighted_num_examples : builtins.float = ...,
        features : typing.Optional[typing.Iterable[global___FeatureNameStatistics]] = ...,
        cross_features : typing.Optional[typing.Iterable[global___CrossFeatureStatistics]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"cross_features",b"cross_features",u"features",b"features",u"name",b"name",u"num_examples",b"num_examples",u"weighted_num_examples",b"weighted_num_examples"]) -> None: ...
global___DatasetFeatureStatistics = DatasetFeatureStatistics

class CrossFeatureStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    PATH_X_FIELD_NUMBER: builtins.int
    PATH_Y_FIELD_NUMBER: builtins.int
    COUNT_FIELD_NUMBER: builtins.int
    NUM_CROSS_STATS_FIELD_NUMBER: builtins.int
    CATEGORICAL_CROSS_STATS_FIELD_NUMBER: builtins.int
    count: builtins.int = ...

    @property
    def path_x(self) -> tensorflow_metadata.proto.v0.path_pb2.Path: ...

    @property
    def path_y(self) -> tensorflow_metadata.proto.v0.path_pb2.Path: ...

    @property
    def num_cross_stats(self) -> global___NumericCrossStatistics: ...

    @property
    def categorical_cross_stats(self) -> global___CategoricalCrossStatistics: ...

    def __init__(self,
        *,
        path_x : typing.Optional[tensorflow_metadata.proto.v0.path_pb2.Path] = ...,
        path_y : typing.Optional[tensorflow_metadata.proto.v0.path_pb2.Path] = ...,
        count : builtins.int = ...,
        num_cross_stats : typing.Optional[global___NumericCrossStatistics] = ...,
        categorical_cross_stats : typing.Optional[global___CategoricalCrossStatistics] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"categorical_cross_stats",b"categorical_cross_stats",u"cross_stats",b"cross_stats",u"num_cross_stats",b"num_cross_stats",u"path_x",b"path_x",u"path_y",b"path_y"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"categorical_cross_stats",b"categorical_cross_stats",u"count",b"count",u"cross_stats",b"cross_stats",u"num_cross_stats",b"num_cross_stats",u"path_x",b"path_x",u"path_y",b"path_y"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"cross_stats",b"cross_stats"]) -> typing_extensions.Literal["num_cross_stats","categorical_cross_stats"]: ...
global___CrossFeatureStatistics = CrossFeatureStatistics

class NumericCrossStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    CORRELATION_FIELD_NUMBER: builtins.int
    COVARIANCE_FIELD_NUMBER: builtins.int
    correlation: builtins.float = ...
    covariance: builtins.float = ...

    def __init__(self,
        *,
        correlation : builtins.float = ...,
        covariance : builtins.float = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"correlation",b"correlation",u"covariance",b"covariance"]) -> None: ...
global___NumericCrossStatistics = NumericCrossStatistics

class CategoricalCrossStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    LIFT_FIELD_NUMBER: builtins.int

    @property
    def lift(self) -> global___LiftStatistics: ...

    def __init__(self,
        *,
        lift : typing.Optional[global___LiftStatistics] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"lift",b"lift"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"lift",b"lift"]) -> None: ...
global___CategoricalCrossStatistics = CategoricalCrossStatistics

class LiftStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    LIFT_SERIES_FIELD_NUMBER: builtins.int
    WEIGHTED_LIFT_SERIES_FIELD_NUMBER: builtins.int

    @property
    def lift_series(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___LiftSeries]: ...

    @property
    def weighted_lift_series(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___LiftSeries]: ...

    def __init__(self,
        *,
        lift_series : typing.Optional[typing.Iterable[global___LiftSeries]] = ...,
        weighted_lift_series : typing.Optional[typing.Iterable[global___LiftSeries]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"lift_series",b"lift_series",u"weighted_lift_series",b"weighted_lift_series"]) -> None: ...
global___LiftStatistics = LiftStatistics

class LiftSeries(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class Bucket(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        LOW_VALUE_FIELD_NUMBER: builtins.int
        HIGH_VALUE_FIELD_NUMBER: builtins.int
        low_value: builtins.float = ...
        high_value: builtins.float = ...

        def __init__(self,
            *,
            low_value : builtins.float = ...,
            high_value : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"high_value",b"high_value",u"low_value",b"low_value"]) -> None: ...

    class LiftValue(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        X_INT_FIELD_NUMBER: builtins.int
        X_STRING_FIELD_NUMBER: builtins.int
        LIFT_FIELD_NUMBER: builtins.int
        X_COUNT_FIELD_NUMBER: builtins.int
        WEIGHTED_X_COUNT_FIELD_NUMBER: builtins.int
        X_AND_Y_COUNT_FIELD_NUMBER: builtins.int
        WEIGHTED_X_AND_Y_COUNT_FIELD_NUMBER: builtins.int
        x_int: builtins.int = ...
        x_string: typing.Text = ...
        lift: builtins.float = ...
        x_count: builtins.int = ...
        weighted_x_count: builtins.float = ...
        x_and_y_count: builtins.int = ...
        weighted_x_and_y_count: builtins.float = ...

        def __init__(self,
            *,
            x_int : builtins.int = ...,
            x_string : typing.Text = ...,
            lift : builtins.float = ...,
            x_count : builtins.int = ...,
            weighted_x_count : builtins.float = ...,
            x_and_y_count : builtins.int = ...,
            weighted_x_and_y_count : builtins.float = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"weighted_x_and_y_count",b"weighted_x_and_y_count",u"weighted_x_count",b"weighted_x_count",u"x_and_y_count",b"x_and_y_count",u"x_and_y_count_value",b"x_and_y_count_value",u"x_count",b"x_count",u"x_count_value",b"x_count_value",u"x_int",b"x_int",u"x_string",b"x_string",u"x_value",b"x_value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"lift",b"lift",u"weighted_x_and_y_count",b"weighted_x_and_y_count",u"weighted_x_count",b"weighted_x_count",u"x_and_y_count",b"x_and_y_count",u"x_and_y_count_value",b"x_and_y_count_value",u"x_count",b"x_count",u"x_count_value",b"x_count_value",u"x_int",b"x_int",u"x_string",b"x_string",u"x_value",b"x_value"]) -> None: ...
        @typing.overload
        def WhichOneof(self, oneof_group: typing_extensions.Literal[u"x_and_y_count_value",b"x_and_y_count_value"]) -> typing_extensions.Literal["x_and_y_count","weighted_x_and_y_count"]: ...
        @typing.overload
        def WhichOneof(self, oneof_group: typing_extensions.Literal[u"x_count_value",b"x_count_value"]) -> typing_extensions.Literal["x_count","weighted_x_count"]: ...
        @typing.overload
        def WhichOneof(self, oneof_group: typing_extensions.Literal[u"x_value",b"x_value"]) -> typing_extensions.Literal["x_int","x_string"]: ...

    Y_INT_FIELD_NUMBER: builtins.int
    Y_STRING_FIELD_NUMBER: builtins.int
    Y_BUCKET_FIELD_NUMBER: builtins.int
    Y_COUNT_FIELD_NUMBER: builtins.int
    WEIGHTED_Y_COUNT_FIELD_NUMBER: builtins.int
    LIFT_VALUES_FIELD_NUMBER: builtins.int
    y_int: builtins.int = ...
    y_string: typing.Text = ...
    y_count: builtins.int = ...
    weighted_y_count: builtins.float = ...

    @property
    def y_bucket(self) -> global___LiftSeries.Bucket: ...

    @property
    def lift_values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___LiftSeries.LiftValue]: ...

    def __init__(self,
        *,
        y_int : builtins.int = ...,
        y_string : typing.Text = ...,
        y_bucket : typing.Optional[global___LiftSeries.Bucket] = ...,
        y_count : builtins.int = ...,
        weighted_y_count : builtins.float = ...,
        lift_values : typing.Optional[typing.Iterable[global___LiftSeries.LiftValue]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"weighted_y_count",b"weighted_y_count",u"y_bucket",b"y_bucket",u"y_count",b"y_count",u"y_count_value",b"y_count_value",u"y_int",b"y_int",u"y_string",b"y_string",u"y_value",b"y_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"lift_values",b"lift_values",u"weighted_y_count",b"weighted_y_count",u"y_bucket",b"y_bucket",u"y_count",b"y_count",u"y_count_value",b"y_count_value",u"y_int",b"y_int",u"y_string",b"y_string",u"y_value",b"y_value"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"y_count_value",b"y_count_value"]) -> typing_extensions.Literal["y_count","weighted_y_count"]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"y_value",b"y_value"]) -> typing_extensions.Literal["y_int","y_string","y_bucket"]: ...
global___LiftSeries = LiftSeries

class FeatureNameStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class _Type(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Type.V], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
        INT = FeatureNameStatistics.Type.V(0)
        FLOAT = FeatureNameStatistics.Type.V(1)
        STRING = FeatureNameStatistics.Type.V(2)
        BYTES = FeatureNameStatistics.Type.V(3)
        STRUCT = FeatureNameStatistics.Type.V(4)
    class Type(metaclass=_Type):
        V = typing.NewType('V', builtins.int)
    INT = FeatureNameStatistics.Type.V(0)
    FLOAT = FeatureNameStatistics.Type.V(1)
    STRING = FeatureNameStatistics.Type.V(2)
    BYTES = FeatureNameStatistics.Type.V(3)
    STRUCT = FeatureNameStatistics.Type.V(4)

    NAME_FIELD_NUMBER: builtins.int
    PATH_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    NUM_STATS_FIELD_NUMBER: builtins.int
    STRING_STATS_FIELD_NUMBER: builtins.int
    BYTES_STATS_FIELD_NUMBER: builtins.int
    STRUCT_STATS_FIELD_NUMBER: builtins.int
    CUSTOM_STATS_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    type: global___FeatureNameStatistics.Type.V = ...

    @property
    def path(self) -> tensorflow_metadata.proto.v0.path_pb2.Path: ...

    @property
    def num_stats(self) -> global___NumericStatistics: ...

    @property
    def string_stats(self) -> global___StringStatistics: ...

    @property
    def bytes_stats(self) -> global___BytesStatistics: ...

    @property
    def struct_stats(self) -> global___StructStatistics: ...

    @property
    def custom_stats(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___CustomStatistic]: ...

    def __init__(self,
        *,
        name : typing.Text = ...,
        path : typing.Optional[tensorflow_metadata.proto.v0.path_pb2.Path] = ...,
        type : global___FeatureNameStatistics.Type.V = ...,
        num_stats : typing.Optional[global___NumericStatistics] = ...,
        string_stats : typing.Optional[global___StringStatistics] = ...,
        bytes_stats : typing.Optional[global___BytesStatistics] = ...,
        struct_stats : typing.Optional[global___StructStatistics] = ...,
        custom_stats : typing.Optional[typing.Iterable[global___CustomStatistic]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"bytes_stats",b"bytes_stats",u"field_id",b"field_id",u"name",b"name",u"num_stats",b"num_stats",u"path",b"path",u"stats",b"stats",u"string_stats",b"string_stats",u"struct_stats",b"struct_stats"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"bytes_stats",b"bytes_stats",u"custom_stats",b"custom_stats",u"field_id",b"field_id",u"name",b"name",u"num_stats",b"num_stats",u"path",b"path",u"stats",b"stats",u"string_stats",b"string_stats",u"struct_stats",b"struct_stats",u"type",b"type"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"field_id",b"field_id"]) -> typing_extensions.Literal["name","path"]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"stats",b"stats"]) -> typing_extensions.Literal["num_stats","string_stats","bytes_stats","struct_stats"]: ...
global___FeatureNameStatistics = FeatureNameStatistics

class WeightedCommonStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NUM_NON_MISSING_FIELD_NUMBER: builtins.int
    NUM_MISSING_FIELD_NUMBER: builtins.int
    AVG_NUM_VALUES_FIELD_NUMBER: builtins.int
    TOT_NUM_VALUES_FIELD_NUMBER: builtins.int
    num_non_missing: builtins.float = ...
    num_missing: builtins.float = ...
    avg_num_values: builtins.float = ...
    tot_num_values: builtins.float = ...

    def __init__(self,
        *,
        num_non_missing : builtins.float = ...,
        num_missing : builtins.float = ...,
        avg_num_values : builtins.float = ...,
        tot_num_values : builtins.float = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"avg_num_values",b"avg_num_values",u"num_missing",b"num_missing",u"num_non_missing",b"num_non_missing",u"tot_num_values",b"tot_num_values"]) -> None: ...
global___WeightedCommonStatistics = WeightedCommonStatistics

class CustomStatistic(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    NUM_FIELD_NUMBER: builtins.int
    STR_FIELD_NUMBER: builtins.int
    HISTOGRAM_FIELD_NUMBER: builtins.int
    RANK_HISTOGRAM_FIELD_NUMBER: builtins.int
    name: typing.Text = ...
    num: builtins.float = ...
    str: typing.Text = ...

    @property
    def histogram(self) -> global___Histogram: ...

    @property
    def rank_histogram(self) -> global___RankHistogram: ...

    def __init__(self,
        *,
        name : typing.Text = ...,
        num : builtins.float = ...,
        str : typing.Text = ...,
        histogram : typing.Optional[global___Histogram] = ...,
        rank_histogram : typing.Optional[global___RankHistogram] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"histogram",b"histogram",u"num",b"num",u"rank_histogram",b"rank_histogram",u"str",b"str",u"val",b"val"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"histogram",b"histogram",u"name",b"name",u"num",b"num",u"rank_histogram",b"rank_histogram",u"str",b"str",u"val",b"val"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"val",b"val"]) -> typing_extensions.Literal["num","str","histogram","rank_histogram"]: ...
global___CustomStatistic = CustomStatistic

class NumericStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    COMMON_STATS_FIELD_NUMBER: builtins.int
    MEAN_FIELD_NUMBER: builtins.int
    STD_DEV_FIELD_NUMBER: builtins.int
    NUM_ZEROS_FIELD_NUMBER: builtins.int
    MIN_FIELD_NUMBER: builtins.int
    MEDIAN_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    HISTOGRAMS_FIELD_NUMBER: builtins.int
    WEIGHTED_NUMERIC_STATS_FIELD_NUMBER: builtins.int
    mean: builtins.float = ...
    std_dev: builtins.float = ...
    num_zeros: builtins.int = ...
    min: builtins.float = ...
    median: builtins.float = ...
    max: builtins.float = ...

    @property
    def common_stats(self) -> global___CommonStatistics: ...

    @property
    def histograms(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Histogram]: ...

    @property
    def weighted_numeric_stats(self) -> global___WeightedNumericStatistics: ...

    def __init__(self,
        *,
        common_stats : typing.Optional[global___CommonStatistics] = ...,
        mean : builtins.float = ...,
        std_dev : builtins.float = ...,
        num_zeros : builtins.int = ...,
        min : builtins.float = ...,
        median : builtins.float = ...,
        max : builtins.float = ...,
        histograms : typing.Optional[typing.Iterable[global___Histogram]] = ...,
        weighted_numeric_stats : typing.Optional[global___WeightedNumericStatistics] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"common_stats",b"common_stats",u"weighted_numeric_stats",b"weighted_numeric_stats"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"common_stats",b"common_stats",u"histograms",b"histograms",u"max",b"max",u"mean",b"mean",u"median",b"median",u"min",b"min",u"num_zeros",b"num_zeros",u"std_dev",b"std_dev",u"weighted_numeric_stats",b"weighted_numeric_stats"]) -> None: ...
global___NumericStatistics = NumericStatistics

class StringStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class FreqAndValue(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        VALUE_FIELD_NUMBER: builtins.int
        FREQUENCY_FIELD_NUMBER: builtins.int
        value: typing.Text = ...
        frequency: builtins.float = ...

        def __init__(self,
            *,
            value : typing.Text = ...,
            frequency : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"frequency",b"frequency",u"value",b"value"]) -> None: ...

    COMMON_STATS_FIELD_NUMBER: builtins.int
    UNIQUE_FIELD_NUMBER: builtins.int
    TOP_VALUES_FIELD_NUMBER: builtins.int
    AVG_LENGTH_FIELD_NUMBER: builtins.int
    RANK_HISTOGRAM_FIELD_NUMBER: builtins.int
    WEIGHTED_STRING_STATS_FIELD_NUMBER: builtins.int
    VOCABULARY_FILE_FIELD_NUMBER: builtins.int
    unique: builtins.int = ...
    avg_length: builtins.float = ...
    vocabulary_file: typing.Text = ...

    @property
    def common_stats(self) -> global___CommonStatistics: ...

    @property
    def top_values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StringStatistics.FreqAndValue]: ...

    @property
    def rank_histogram(self) -> global___RankHistogram: ...

    @property
    def weighted_string_stats(self) -> global___WeightedStringStatistics: ...

    def __init__(self,
        *,
        common_stats : typing.Optional[global___CommonStatistics] = ...,
        unique : builtins.int = ...,
        top_values : typing.Optional[typing.Iterable[global___StringStatistics.FreqAndValue]] = ...,
        avg_length : builtins.float = ...,
        rank_histogram : typing.Optional[global___RankHistogram] = ...,
        weighted_string_stats : typing.Optional[global___WeightedStringStatistics] = ...,
        vocabulary_file : typing.Text = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"common_stats",b"common_stats",u"rank_histogram",b"rank_histogram",u"weighted_string_stats",b"weighted_string_stats"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"avg_length",b"avg_length",u"common_stats",b"common_stats",u"rank_histogram",b"rank_histogram",u"top_values",b"top_values",u"unique",b"unique",u"vocabulary_file",b"vocabulary_file",u"weighted_string_stats",b"weighted_string_stats"]) -> None: ...
global___StringStatistics = StringStatistics

class WeightedNumericStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    MEAN_FIELD_NUMBER: builtins.int
    STD_DEV_FIELD_NUMBER: builtins.int
    MEDIAN_FIELD_NUMBER: builtins.int
    HISTOGRAMS_FIELD_NUMBER: builtins.int
    mean: builtins.float = ...
    std_dev: builtins.float = ...
    median: builtins.float = ...

    @property
    def histograms(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Histogram]: ...

    def __init__(self,
        *,
        mean : builtins.float = ...,
        std_dev : builtins.float = ...,
        median : builtins.float = ...,
        histograms : typing.Optional[typing.Iterable[global___Histogram]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"histograms",b"histograms",u"mean",b"mean",u"median",b"median",u"std_dev",b"std_dev"]) -> None: ...
global___WeightedNumericStatistics = WeightedNumericStatistics

class WeightedStringStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    TOP_VALUES_FIELD_NUMBER: builtins.int
    RANK_HISTOGRAM_FIELD_NUMBER: builtins.int

    @property
    def top_values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StringStatistics.FreqAndValue]: ...

    @property
    def rank_histogram(self) -> global___RankHistogram: ...

    def __init__(self,
        *,
        top_values : typing.Optional[typing.Iterable[global___StringStatistics.FreqAndValue]] = ...,
        rank_histogram : typing.Optional[global___RankHistogram] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"rank_histogram",b"rank_histogram"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"rank_histogram",b"rank_histogram",u"top_values",b"top_values"]) -> None: ...
global___WeightedStringStatistics = WeightedStringStatistics

class BytesStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    COMMON_STATS_FIELD_NUMBER: builtins.int
    UNIQUE_FIELD_NUMBER: builtins.int
    AVG_NUM_BYTES_FIELD_NUMBER: builtins.int
    MIN_NUM_BYTES_FIELD_NUMBER: builtins.int
    MAX_NUM_BYTES_FIELD_NUMBER: builtins.int
    unique: builtins.int = ...
    avg_num_bytes: builtins.float = ...
    min_num_bytes: builtins.float = ...
    max_num_bytes: builtins.float = ...

    @property
    def common_stats(self) -> global___CommonStatistics: ...

    def __init__(self,
        *,
        common_stats : typing.Optional[global___CommonStatistics] = ...,
        unique : builtins.int = ...,
        avg_num_bytes : builtins.float = ...,
        min_num_bytes : builtins.float = ...,
        max_num_bytes : builtins.float = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"common_stats",b"common_stats"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"avg_num_bytes",b"avg_num_bytes",u"common_stats",b"common_stats",u"max_num_bytes",b"max_num_bytes",u"min_num_bytes",b"min_num_bytes",u"unique",b"unique"]) -> None: ...
global___BytesStatistics = BytesStatistics

class StructStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    COMMON_STATS_FIELD_NUMBER: builtins.int

    @property
    def common_stats(self) -> global___CommonStatistics: ...

    def __init__(self,
        *,
        common_stats : typing.Optional[global___CommonStatistics] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"common_stats",b"common_stats"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"common_stats",b"common_stats"]) -> None: ...
global___StructStatistics = StructStatistics

class CommonStatistics(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NUM_NON_MISSING_FIELD_NUMBER: builtins.int
    NUM_MISSING_FIELD_NUMBER: builtins.int
    MIN_NUM_VALUES_FIELD_NUMBER: builtins.int
    MAX_NUM_VALUES_FIELD_NUMBER: builtins.int
    AVG_NUM_VALUES_FIELD_NUMBER: builtins.int
    TOT_NUM_VALUES_FIELD_NUMBER: builtins.int
    NUM_VALUES_HISTOGRAM_FIELD_NUMBER: builtins.int
    WEIGHTED_COMMON_STATS_FIELD_NUMBER: builtins.int
    FEATURE_LIST_LENGTH_HISTOGRAM_FIELD_NUMBER: builtins.int
    num_non_missing: builtins.int = ...
    num_missing: builtins.int = ...
    min_num_values: builtins.int = ...
    max_num_values: builtins.int = ...
    avg_num_values: builtins.float = ...
    tot_num_values: builtins.int = ...

    @property
    def num_values_histogram(self) -> global___Histogram: ...

    @property
    def weighted_common_stats(self) -> global___WeightedCommonStatistics: ...

    @property
    def feature_list_length_histogram(self) -> global___Histogram: ...

    def __init__(self,
        *,
        num_non_missing : builtins.int = ...,
        num_missing : builtins.int = ...,
        min_num_values : builtins.int = ...,
        max_num_values : builtins.int = ...,
        avg_num_values : builtins.float = ...,
        tot_num_values : builtins.int = ...,
        num_values_histogram : typing.Optional[global___Histogram] = ...,
        weighted_common_stats : typing.Optional[global___WeightedCommonStatistics] = ...,
        feature_list_length_histogram : typing.Optional[global___Histogram] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"feature_list_length_histogram",b"feature_list_length_histogram",u"num_values_histogram",b"num_values_histogram",u"weighted_common_stats",b"weighted_common_stats"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"avg_num_values",b"avg_num_values",u"feature_list_length_histogram",b"feature_list_length_histogram",u"max_num_values",b"max_num_values",u"min_num_values",b"min_num_values",u"num_missing",b"num_missing",u"num_non_missing",b"num_non_missing",u"num_values_histogram",b"num_values_histogram",u"tot_num_values",b"tot_num_values",u"weighted_common_stats",b"weighted_common_stats"]) -> None: ...
global___CommonStatistics = CommonStatistics

class Histogram(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class _HistogramType(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[HistogramType.V], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
        STANDARD = Histogram.HistogramType.V(0)
        QUANTILES = Histogram.HistogramType.V(1)
    class HistogramType(metaclass=_HistogramType):
        V = typing.NewType('V', builtins.int)
    STANDARD = Histogram.HistogramType.V(0)
    QUANTILES = Histogram.HistogramType.V(1)

    class Bucket(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        LOW_VALUE_FIELD_NUMBER: builtins.int
        HIGH_VALUE_FIELD_NUMBER: builtins.int
        SAMPLE_COUNT_FIELD_NUMBER: builtins.int
        low_value: builtins.float = ...
        high_value: builtins.float = ...
        sample_count: builtins.float = ...

        def __init__(self,
            *,
            low_value : builtins.float = ...,
            high_value : builtins.float = ...,
            sample_count : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"high_value",b"high_value",u"low_value",b"low_value",u"sample_count",b"sample_count"]) -> None: ...

    NUM_NAN_FIELD_NUMBER: builtins.int
    NUM_UNDEFINED_FIELD_NUMBER: builtins.int
    BUCKETS_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    num_nan: builtins.int = ...
    num_undefined: builtins.int = ...
    type: global___Histogram.HistogramType.V = ...
    name: typing.Text = ...

    @property
    def buckets(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Histogram.Bucket]: ...

    def __init__(self,
        *,
        num_nan : builtins.int = ...,
        num_undefined : builtins.int = ...,
        buckets : typing.Optional[typing.Iterable[global___Histogram.Bucket]] = ...,
        type : global___Histogram.HistogramType.V = ...,
        name : typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"buckets",b"buckets",u"name",b"name",u"num_nan",b"num_nan",u"num_undefined",b"num_undefined",u"type",b"type"]) -> None: ...
global___Histogram = Histogram

class RankHistogram(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class Bucket(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        LOW_RANK_FIELD_NUMBER: builtins.int
        HIGH_RANK_FIELD_NUMBER: builtins.int
        LABEL_FIELD_NUMBER: builtins.int
        SAMPLE_COUNT_FIELD_NUMBER: builtins.int
        low_rank: builtins.int = ...
        high_rank: builtins.int = ...
        label: typing.Text = ...
        sample_count: builtins.float = ...

        def __init__(self,
            *,
            low_rank : builtins.int = ...,
            high_rank : builtins.int = ...,
            label : typing.Text = ...,
            sample_count : builtins.float = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"high_rank",b"high_rank",u"label",b"label",u"low_rank",b"low_rank",u"sample_count",b"sample_count"]) -> None: ...

    BUCKETS_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    name: typing.Text = ...

    @property
    def buckets(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___RankHistogram.Bucket]: ...

    def __init__(self,
        *,
        buckets : typing.Optional[typing.Iterable[global___RankHistogram.Bucket]] = ...,
        name : typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"buckets",b"buckets",u"name",b"name"]) -> None: ...
global___RankHistogram = RankHistogram



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Tuple, Union

from google.protobuf.duration_pb2 import Duration
from google.protobuf.json_format import MessageToJson
from google.protobuf.timestamp_pb2 import Timestamp

from feast import utils
from feast.data_source import BigQuerySource, DataSource, FileSource
from feast.feature import Feature
from feast.protos.feast.core.FeatureView_pb2 import FeatureView as FeatureViewProto
from feast.protos.feast.core.FeatureView_pb2 import (
    FeatureViewMeta as FeatureViewMetaProto,
)
from feast.protos.feast.core.FeatureView_pb2 import (
    FeatureViewSpec as FeatureViewSpecProto,
)
from feast.protos.feast.core.FeatureView_pb2 import (
    MaterializationInterval as MaterializationIntervalProto,
)
from feast.value_type import ValueType


class FeatureView:
    """
    A FeatureView defines a logical grouping of serveable features.
    """

    name: str
    entities: List[str]
    features: List[Feature]
    tags: Optional[Dict[str, str]]
    ttl: Optional[timedelta]
    online: bool
    input: Union[BigQuerySource, FileSource]

    created_timestamp: Optional[Timestamp] = None
    last_updated_timestamp: Optional[Timestamp] = None
    materialization_intervals: List[Tuple[datetime, datetime]]

    def __init__(
        self,
        name: str,
        entities: List[str],
        features: List[Feature],
        ttl: Optional[Union[Duration, timedelta]],
        input: Union[BigQuerySource, FileSource],
        tags: Optional[Dict[str, str]] = None,
        online: bool = True,
    ):
        cols = [entity for entity in entities] + [feat.name for feat in features]
        for col in cols:
            if input.field_mapping is not None and col in input.field_mapping.keys():
                raise ValueError(
                    f"The field {col} is mapped to {input.field_mapping[col]} for this data source. Please either remove this field mapping or use {input.field_mapping[col]} as the Entity or Feature name."
                )

        self.name = name
        self.entities = entities
        self.features = features
        self.tags = tags

        if isinstance(ttl, Duration):
            self.ttl = timedelta(seconds=int(ttl.seconds))
        else:
            self.ttl = ttl

        self.online = online
        self.input = input

        self.materialization_intervals = []

    def __repr__(self):
        items = (f"{k} = {v}" for k, v in self.__dict__.items())
        return f"<{self.__class__.__name__}({', '.join(items)})>"

    def __str__(self):
        return str(MessageToJson(self.to_proto()))

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if not isinstance(other, FeatureView):
            raise TypeError(
                "Comparisons should only involve FeatureView class objects."
            )

        if (
            self.tags != other.tags
            or self.name != other.name
            or self.ttl != other.ttl
            or self.online != other.online
        ):
            return False

        if sorted(self.entities) != sorted(other.entities):
            return False
        if sorted(self.features) != sorted(other.features):
            return False
        if self.input != other.input:
            return False

        return True

    def is_valid(self):
        """
        Validates the state of a feature view locally. Raises an exception
        if feature view is invalid.
        """

        if not self.name:
            raise ValueError("Feature view needs a name")

        if not self.entities:
            raise ValueError("Feature view has no entities")

    def to_proto(self) -> FeatureViewProto:
        """
        Converts an feature view object to its protobuf representation.

        Returns:
            FeatureViewProto protobuf
        """

        meta = FeatureViewMetaProto(
            created_timestamp=self.created_timestamp,
            last_updated_timestamp=self.last_updated_timestamp,
            materialization_intervals=[],
        )
        for interval in self.materialization_intervals:
            interval_proto = MaterializationIntervalProto()
            interval_proto.start_time.FromDatetime(interval[0])
            interval_proto.end_time.FromDatetime(interval[1])
            meta.materialization_intervals.append(interval_proto)

        if self.ttl is not None:
            ttl_duration = Duration()
            ttl_duration.FromTimedelta(self.ttl)

        spec = FeatureViewSpecProto(
            name=self.name,
            entities=self.entities,
            features=[feature.to_proto() for feature in self.features],
            tags=self.tags,
            ttl=(ttl_duration if ttl_duration is not None else None),
            online=self.online,
            input=self.input.to_proto(),
        )

        return FeatureViewProto(spec=spec, meta=meta)

    @classmethod
    def from_proto(cls, feature_view_proto: FeatureViewProto):
        """
        Creates a feature view from a protobuf representation of a feature view

        Args:
            feature_view_proto: A protobuf representation of a feature view

        Returns:
            Returns a FeatureViewProto object based on the feature view protobuf
        """

        feature_view = cls(
            name=feature_view_proto.spec.name,
            entities=[entity for entity in feature_view_proto.spec.entities],
            features=[
                Feature(
                    name=feature.name,
                    dtype=ValueType(feature.value_type),
                    labels=feature.labels,
                )
                for feature in feature_view_proto.spec.features
            ],
            tags=dict(feature_view_proto.spec.tags),
            online=feature_view_proto.spec.online,
            ttl=(
                None
                if feature_view_proto.spec.ttl.seconds == 0
                and feature_view_proto.spec.ttl.nanos == 0
                else feature_view_proto.spec.ttl
            ),
            input=DataSource.from_proto(feature_view_proto.spec.input),
        )

        feature_view.created_timestamp = feature_view_proto.meta.created_timestamp

        for interval in feature_view_proto.meta.materialization_intervals:
            feature_view.materialization_intervals.append(
                (
                    utils.make_tzaware(interval.start_time.ToDatetime()),
                    utils.make_tzaware(interval.end_time.ToDatetime()),
                )
            )

        return feature_view

    @property
    def most_recent_end_time(self) -> Optional[datetime]:
        if len(self.materialization_intervals) == 0:
            return None
        return max([interval[1] for interval in self.materialization_intervals])



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from pathlib import Path
from tempfile import TemporaryFile
from typing import Callable, List, Optional
from urllib.parse import urlparse

from google.auth.exceptions import DefaultCredentialsError

from feast.entity import Entity
from feast.errors import (
    EntityNotFoundException,
    FeatureTableNotFoundException,
    FeatureViewNotFoundException,
)
from feast.feature_table import FeatureTable
from feast.feature_view import FeatureView
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto

REGISTRY_SCHEMA_VERSION = "1"


class Registry:
    """
    Registry: A registry allows for the management and persistence of feature definitions and related metadata.
    """

    cached_registry_proto: Optional[RegistryProto] = None
    cached_registry_proto_created: Optional[datetime] = None
    cached_registry_proto_ttl: timedelta
    cache_being_updated: bool = False

    def __init__(self, registry_path: str, repo_path: Path, cache_ttl: timedelta):
        """
        Create the Registry object.

        Args:
            repo_path: Path to the base of the Feast repository
            cache_ttl: The amount of time that cached registry state stays valid
            registry_path: filepath or GCS URI that is the location of the object store registry,
            or where it will be created if it does not exist yet.
        """
        uri = urlparse(registry_path)
        if uri.scheme == "gs":
            self._registry_store: RegistryStore = GCSRegistryStore(registry_path)
        elif uri.scheme == "file" or uri.scheme == "":
            self._registry_store = LocalRegistryStore(
                repo_path=repo_path, registry_path_string=registry_path
            )
        else:
            raise Exception(
                f"Registry path {registry_path} has unsupported scheme {uri.scheme}. Supported schemes are file and gs."
            )
        self.cached_registry_proto_ttl = cache_ttl
        return

    def _initialize_registry(self):
        """Explicitly forces the initialization of a registry"""
        self._registry_store.update_registry_proto()

    def apply_entity(self, entity: Entity, project: str):
        """
        Registers a single entity with Feast

        Args:
            entity: Entity that will be registered
            project: Feast project that this entity belongs to
        """
        entity.is_valid()
        entity_proto = entity.to_proto()
        entity_proto.spec.project = project

        def updater(registry_proto: RegistryProto):
            for idx, existing_entity_proto in enumerate(registry_proto.entities):
                if (
                    existing_entity_proto.spec.name == entity_proto.spec.name
                    and existing_entity_proto.spec.project == project
                ):
                    del registry_proto.entities[idx]
                    registry_proto.entities.append(entity_proto)
                    return registry_proto
            registry_proto.entities.append(entity_proto)
            return registry_proto

        self._registry_store.update_registry_proto(updater)
        return

    def list_entities(self, project: str, allow_cache: bool = False) -> List[Entity]:
        """
        Retrieve a list of entities from the registry

        Args:
            allow_cache: Whether to allow returning entities from a cached registry
            project: Filter entities based on project name

        Returns:
            List of entities
        """
        registry_proto = self._get_registry_proto(allow_cache=allow_cache)
        entities = []
        for entity_proto in registry_proto.entities:
            if entity_proto.spec.project == project:
                entities.append(Entity.from_proto(entity_proto))
        return entities

    def get_entity(self, name: str, project: str, allow_cache: bool = False) -> Entity:
        """
        Retrieves an entity.

        Args:
            name: Name of entity
            project: Feast project that this entity belongs to

        Returns:
            Returns either the specified entity, or raises an exception if
            none is found
        """
        registry_proto = self._get_registry_proto(allow_cache=allow_cache)
        for entity_proto in registry_proto.entities:
            if entity_proto.spec.name == name and entity_proto.spec.project == project:
                return Entity.from_proto(entity_proto)
        raise EntityNotFoundException(project, name)

    def apply_feature_table(self, feature_table: FeatureTable, project: str):
        """
        Registers a single feature table with Feast

        Args:
            feature_table: Feature table that will be registered
            project: Feast project that this feature table belongs to
        """
        feature_table.is_valid()
        feature_table_proto = feature_table.to_proto()
        feature_table_proto.spec.project = project

        def updater(registry_proto: RegistryProto):
            for idx, existing_feature_table_proto in enumerate(
                registry_proto.feature_tables
            ):
                if (
                    existing_feature_table_proto.spec.name
                    == feature_table_proto.spec.name
                    and existing_feature_table_proto.spec.project == project
                ):
                    del registry_proto.feature_tables[idx]
                    registry_proto.feature_tables.append(feature_table_proto)
                    return registry_proto
            registry_proto.feature_tables.append(feature_table_proto)
            return registry_proto

        self._registry_store.update_registry_proto(updater)
        return

    def apply_feature_view(self, feature_view: FeatureView, project: str):
        """
        Registers a single feature view with Feast

        Args:
            feature_view: Feature view that will be registered
            project: Feast project that this feature view belongs to
        """
        feature_view.is_valid()
        feature_view_proto = feature_view.to_proto()
        feature_view_proto.spec.project = project

        def updater(registry_proto: RegistryProto):
            for idx, existing_feature_view_proto in enumerate(
                registry_proto.feature_views
            ):
                if (
                    existing_feature_view_proto.spec.name
                    == feature_view_proto.spec.name
                    and existing_feature_view_proto.spec.project == project
                ):
                    del registry_proto.feature_views[idx]
                    registry_proto.feature_views.append(feature_view_proto)
                    return registry_proto
            registry_proto.feature_views.append(feature_view_proto)
            return registry_proto

        self._registry_store.update_registry_proto(updater)

    def list_feature_tables(self, project: str) -> List[FeatureTable]:
        """
        Retrieve a list of feature tables from the registry

        Args:
            project: Filter feature tables based on project name

        Returns:
            List of feature tables
        """
        registry_proto = self._get_registry_proto()
        feature_tables = []
        for feature_table_proto in registry_proto.feature_tables:
            if feature_table_proto.spec.project == project:
                feature_tables.append(FeatureTable.from_proto(feature_table_proto))
        return feature_tables

    def list_feature_views(
        self, project: str, allow_cache: bool = False
    ) -> List[FeatureView]:
        """
        Retrieve a list of feature views from the registry

        Args:
            allow_cache: Allow returning feature views from the cached registry
            project: Filter feature tables based on project name

        Returns:
            List of feature views
        """
        registry_proto = self._get_registry_proto(allow_cache=allow_cache)
        feature_views = []
        for feature_view_proto in registry_proto.feature_views:
            if feature_view_proto.spec.project == project:
                feature_views.append(FeatureView.from_proto(feature_view_proto))
        return feature_views

    def get_feature_table(self, name: str, project: str) -> FeatureTable:
        """
        Retrieves a feature table.

        Args:
            name: Name of feature table
            project: Feast project that this feature table belongs to

        Returns:
            Returns either the specified feature table, or raises an exception if
            none is found
        """
        registry_proto = self._get_registry_proto()
        for feature_table_proto in registry_proto.feature_tables:
            if (
                feature_table_proto.spec.name == name
                and feature_table_proto.spec.project == project
            ):
                return FeatureTable.from_proto(feature_table_proto)
        raise FeatureTableNotFoundException(project, name)

    def get_feature_view(self, name: str, project: str) -> FeatureView:
        """
        Retrieves a feature view.

        Args:
            name: Name of feature view
            project: Feast project that this feature view belongs to

        Returns:
            Returns either the specified feature view, or raises an exception if
            none is found
        """
        registry_proto = self._get_registry_proto()
        for feature_view_proto in registry_proto.feature_views:
            if (
                feature_view_proto.spec.name == name
                and feature_view_proto.spec.project == project
            ):
                return FeatureView.from_proto(feature_view_proto)
        raise FeatureViewNotFoundException(name, project)

    def delete_feature_table(self, name: str, project: str):
        """
        Deletes a feature table or raises an exception if not found.

        Args:
            name: Name of feature table
            project: Feast project that this feature table belongs to
        """

        def updater(registry_proto: RegistryProto):
            for idx, existing_feature_table_proto in enumerate(
                registry_proto.feature_tables
            ):
                if (
                    existing_feature_table_proto.spec.name == name
                    and existing_feature_table_proto.spec.project == project
                ):
                    del registry_proto.feature_tables[idx]
                    return registry_proto
            raise FeatureTableNotFoundException(project, name)

        self._registry_store.update_registry_proto(updater)
        return

    def delete_feature_view(self, name: str, project: str):
        """
        Deletes a feature view or raises an exception if not found.

        Args:
            name: Name of feature view
            project: Feast project that this feature view belongs to
        """

        def updater(registry_proto: RegistryProto):
            for idx, existing_feature_view_proto in enumerate(
                registry_proto.feature_views
            ):
                if (
                    existing_feature_view_proto.spec.name == name
                    and existing_feature_view_proto.spec.project == project
                ):
                    del registry_proto.feature_views[idx]
                    return registry_proto
            raise FeatureViewNotFoundException(name, project)

        self._registry_store.update_registry_proto(updater)

    def refresh(self):
        """Refreshes the state of the registry cache by fetching the registry state from the remote registry store."""
        self._get_registry_proto(allow_cache=False)

    def _get_registry_proto(self, allow_cache: bool = False) -> RegistryProto:
        """Returns the cached or remote registry state

        Args:
            allow_cache: Whether to allow the use of the registry cache when fetching the RegistryProto

        Returns: Returns a RegistryProto object which represents the state of the registry
        """
        expired = (
            self.cached_registry_proto is None
            or self.cached_registry_proto_created is None
        ) or (
            self.cached_registry_proto_ttl.total_seconds() > 0  # 0 ttl means infinity
            and (
                datetime.now()
                > (self.cached_registry_proto_created + self.cached_registry_proto_ttl)
            )
        )
        if allow_cache and (not expired or self.cache_being_updated):
            assert isinstance(self.cached_registry_proto, RegistryProto)
            return self.cached_registry_proto

        try:
            self.cache_being_updated = True
            registry_proto = self._registry_store.get_registry_proto()
            self.cached_registry_proto = registry_proto
            self.cached_registry_proto_created = datetime.now()
        except Exception as e:
            raise e
        finally:
            self.cache_being_updated = False
        return registry_proto


class RegistryStore(ABC):
    """
    RegistryStore: abstract base class implemented by specific backends (local file system, GCS)
    containing lower level methods used by the Registry class that are backend-specific.
    """

    @abstractmethod
    def get_registry_proto(self):
        """
        Retrieves the registry proto from the registry path. If there is no file at that path,
        returns an empty registry proto.

        Returns:
            Returns either the registry proto stored at the registry path, or an empty registry proto.
        """
        pass

    @abstractmethod
    def update_registry_proto(self, updater: Callable[[RegistryProto], RegistryProto]):
        """
        Updates the registry using the function passed in. If the registry proto has not been created yet
        this method will create it. This method writes to the registry path.

        Args:
            updater: function that takes in the current registry proto and outputs the desired registry proto
        """
        pass


class LocalRegistryStore(RegistryStore):
    def __init__(self, repo_path: Path, registry_path_string: str):
        registry_path = Path(registry_path_string)
        if registry_path.is_absolute():
            self._filepath = registry_path
        else:
            self._filepath = repo_path.joinpath(registry_path)

    def get_registry_proto(self):
        registry_proto = RegistryProto()
        if self._filepath.exists():
            registry_proto.ParseFromString(self._filepath.read_bytes())
            return registry_proto
        raise FileNotFoundError(
            f'Registry not found at path "{self._filepath}". Have you run "feast apply"?'
        )

    def update_registry_proto(
        self, updater: Callable[[RegistryProto], RegistryProto] = None
    ):
        try:
            registry_proto = self.get_registry_proto()
        except FileNotFoundError:
            registry_proto = RegistryProto()
            registry_proto.registry_schema_version = REGISTRY_SCHEMA_VERSION

        if updater:
            registry_proto = updater(registry_proto)
        self._write_registry(registry_proto)
        return

    def _write_registry(self, registry_proto: RegistryProto):
        registry_proto.version_id = str(uuid.uuid4())
        registry_proto.last_updated.FromDatetime(datetime.utcnow())
        file_dir = self._filepath.parent
        file_dir.mkdir(exist_ok=True)
        self._filepath.write_bytes(registry_proto.SerializeToString())
        return


class GCSRegistryStore(RegistryStore):
    def __init__(self, uri: str):
        try:
            from google.cloud import storage
        except ImportError:
            # TODO: Ensure versioning depends on requirements.txt/setup.py and isn't hardcoded
            raise ImportError(
                "Install package google-cloud-storage==1.20.* for gcs support"
                "run ```pip install google-cloud-storage==1.20.*```"
            )
        try:
            self.gcs_client = storage.Client()
        except DefaultCredentialsError:
            self.gcs_client = storage.Client.create_anonymous_client()
        self._uri = urlparse(uri)
        self._bucket = self._uri.hostname
        self._blob = self._uri.path.lstrip("/")
        return

    def get_registry_proto(self):
        from google.cloud import storage
        from google.cloud.exceptions import NotFound

        file_obj = TemporaryFile()
        registry_proto = RegistryProto()
        try:
            bucket = self.gcs_client.get_bucket(self._bucket)
        except NotFound:
            raise Exception(
                f"No bucket named {self._bucket} exists; please create it first."
            )
        if storage.Blob(bucket=bucket, name=self._blob).exists(self.gcs_client):
            self.gcs_client.download_blob_to_file(
                self._uri.geturl(), file_obj, timeout=30
            )
            file_obj.seek(0)
            registry_proto.ParseFromString(file_obj.read())
            return registry_proto
        raise FileNotFoundError(
            f'Registry not found at path "{self._uri.geturl()}". Have you run "feast apply"?'
        )

    def update_registry_proto(self, updater: Callable[[RegistryProto], RegistryProto]):
        try:
            registry_proto = self.get_registry_proto()
        except FileNotFoundError:
            registry_proto = RegistryProto()
            registry_proto.registry_schema_version = REGISTRY_SCHEMA_VERSION
        registry_proto = updater(registry_proto)
        self._write_registry(registry_proto)
        return

    def _write_registry(self, registry_proto: RegistryProto):
        registry_proto.version_id = str(uuid.uuid4())
        registry_proto.last_updated.FromDatetime(datetime.utcnow())
        # we have already checked the bucket exists so no need to do it again
        gs_bucket = self.gcs_client.get_bucket(self._bucket)
        blob = gs_bucket.blob(self._blob)
        file_obj = TemporaryFile()
        file_obj.write(registry_proto.SerializeToString())
        file_obj.seek(0)
        blob.upload_from_file(file_obj)
        return



#
# Copyright 2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import os
import re
import shutil
from abc import ABC, ABCMeta, abstractmethod
from tempfile import TemporaryFile
from typing import List, Optional, Tuple
from typing.io import IO
from urllib.parse import ParseResult, urlparse

from google.auth.exceptions import DefaultCredentialsError

from feast.config import Config
from feast.constants import ConfigOptions as opt

GS = "gs"
S3 = "s3"
S3A = "s3a"
AZURE_SCHEME = "wasbs"
LOCAL_FILE = "file"


def _hash_fileobj(fileobj: IO[bytes]) -> str:
    """ Compute sha256 hash of a file. File pointer will be reset to 0 on return. """
    fileobj.seek(0)
    h = hashlib.sha256()
    for block in iter(lambda: fileobj.read(2 ** 20), b""):
        h.update(block)
    fileobj.seek(0)
    return h.hexdigest()


def _gen_remote_uri(
    fileobj: IO[bytes],
    remote_uri: Optional[ParseResult],
    remote_path_prefix: Optional[str],
    remote_path_suffix: Optional[str],
    sha256sum: Optional[str],
) -> ParseResult:
    if remote_uri is None:
        assert remote_path_prefix is not None and remote_path_suffix is not None

        if sha256sum is None:
            sha256sum = _hash_fileobj(fileobj)

        return urlparse(
            os.path.join(remote_path_prefix, f"{sha256sum}{remote_path_suffix}")
        )
    else:
        return remote_uri


class AbstractStagingClient(ABC):
    """
    Client used to stage files in order to upload or download datasets into a historical store.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def download_file(self, uri: ParseResult) -> IO[bytes]:
        """
        Downloads a file from an object store and returns a TemporaryFile object
        """
        pass

    @abstractmethod
    def list_files(self, uri: ParseResult) -> List[str]:
        """
        Lists all the files under a directory in an object store.
        """
        pass

    @abstractmethod
    def upload_fileobj(
        self,
        fileobj: IO[bytes],
        local_path: str,
        *,
        remote_uri: Optional[ParseResult] = None,
        remote_path_prefix: Optional[str] = None,
        remote_path_suffix: Optional[str] = None,
    ) -> ParseResult:
        """
        Uploads a file to an object store. You can either specify the destination object URI,
        or destination suffix+prefix. In the latter case, this interface will work as a
        content-addressable storage and the remote path will be computed using sha256 of the
        uploaded content as `$remote_path_prefix/$sha256$remote_path_suffix`

        Args:
            fileobj (IO[bytes]): file-like object containing the data to be uploaded. It needs to
                supports seek() operation in addition to read/write.
            local_path (str): a file name associated with fileobj. This param is only used for
                diagnostic messages. If `fileobj` is a local file, pass its filename here.
            remote_uri (ParseResult or None): destination object URI to upload to
            remote_path_prefix (str or None): destination path prefix to upload to when using
                content-addressable storage mode
            remote_path_suffix (str or None): destination path suffix to upload to when using
                content-addressable storage mode

        Returns:
            ParseResult: the URI to the uploaded file. It would be the same as `remote_uri` if
            `remote_uri` was passed in. Otherwise it will be the path computed from
            `remote_path_prefix` and `remote_path_suffix`.
        """
        pass


class GCSClient(AbstractStagingClient):
    """
    Implementation of AbstractStagingClient for google cloud storage
    """

    def __init__(self):
        try:
            from google.cloud import storage
        except ImportError:
            raise ImportError(
                "Install package google-cloud-storage==1.20.* for gcs staging support"
                "run ```pip install google-cloud-storage==1.20.*```"
            )
        try:
            self.gcs_client = storage.Client(project=None)
        except DefaultCredentialsError:
            self.gcs_client = storage.Client.create_anonymous_client()

    def download_file(self, uri: ParseResult) -> IO[bytes]:
        """
        Downloads a file from google cloud storage and returns a TemporaryFile object

        Args:
            uri (urllib.parse.ParseResult): Parsed uri of the file ex: urlparse("gs://bucket/file.avro")

        Returns:
             TemporaryFile object
        """
        url = uri.geturl()
        file_obj = TemporaryFile()
        self.gcs_client.download_blob_to_file(url, file_obj)
        file_obj.seek(0)
        return file_obj

    def list_files(self, uri: ParseResult) -> List[str]:
        """
        Lists all the files under a directory in google cloud storage if path has wildcard(*) character.

        Args:
            uri (urllib.parse.ParseResult): Parsed uri of this location

        Returns:
            List[str]: A list containing the full path to the file(s) in the
                    remote staging location.
        """

        bucket, path = self._uri_to_bucket_key(uri)
        gs_bucket = self.gcs_client.get_bucket(bucket)

        if "*" in path:
            regex = re.compile(path.replace("*", ".*?").strip("/"))
            blob_list = gs_bucket.list_blobs(
                prefix=path.strip("/").split("*")[0], delimiter="/"
            )
            # File path should not be in path (file path must be longer than path)
            return [
                f"{GS}://{bucket}/{file}"
                for file in [x.name for x in blob_list]
                if re.match(regex, file) and file not in path
            ]
        else:
            return [f"{GS}://{bucket}/{path}"]

    def _uri_to_bucket_key(self, remote_path: ParseResult) -> Tuple[str, str]:
        assert remote_path.hostname is not None
        return remote_path.hostname, remote_path.path.lstrip("/")

    def upload_fileobj(
        self,
        fileobj: IO[bytes],
        local_path: str,
        *,
        remote_uri: Optional[ParseResult] = None,
        remote_path_prefix: Optional[str] = None,
        remote_path_suffix: Optional[str] = None,
    ) -> ParseResult:
        remote_uri = _gen_remote_uri(
            fileobj, remote_uri, remote_path_prefix, remote_path_suffix, None
        )
        bucket, key = self._uri_to_bucket_key(remote_uri)
        gs_bucket = self.gcs_client.get_bucket(bucket)
        blob = gs_bucket.blob(key)
        blob.upload_from_file(fileobj)
        return remote_uri


class S3Client(AbstractStagingClient):
    """
       Implementation of AbstractStagingClient for Aws S3 storage
    """

    def __init__(self, endpoint_url: str = None, url_scheme="s3"):
        try:
            import boto3
        except ImportError:
            raise ImportError(
                "Install package boto3 for s3 staging support"
                "run ```pip install boto3```"
            )
        self.s3_client = boto3.client("s3", endpoint_url=endpoint_url)
        self.url_scheme = url_scheme

    def download_file(self, uri: ParseResult) -> IO[bytes]:
        """
        Downloads a file from AWS s3 storage and returns a TemporaryFile object

        Args:
            uri (urllib.parse.ParseResult): Parsed uri of the file ex: urlparse("s3://bucket/file.avro")
        Returns:
            TemporaryFile object
        """
        bucket, url = self._uri_to_bucket_key(uri)
        file_obj = TemporaryFile()
        self.s3_client.download_fileobj(bucket, url, file_obj)
        return file_obj

    def list_files(self, uri: ParseResult) -> List[str]:
        """
        Lists all the files under a directory in s3 if path has wildcard(*) character.

        Args:
            uri (urllib.parse.ParseResult): Parsed uri of this location

        Returns:
            List[str]: A list containing the full path to the file(s) in the
                    remote staging location.
        """

        bucket, path = self._uri_to_bucket_key(uri)
        if "*" in path:
            regex = re.compile(path.replace("*", ".*?").strip("/"))
            blob_list = self.s3_client.list_objects(
                Bucket=bucket, Prefix=path.strip("/").split("*")[0], Delimiter="/"
            )
            # File path should not be in path (file path must be longer than path)
            return [
                f"{self.url_scheme}://{bucket}/{file}"
                for file in [x["Key"] for x in blob_list["Contents"]]
                if re.match(regex, file) and file not in path
            ]
        else:
            return [f"{self.url_scheme}://{bucket}/{path}"]

    def _uri_to_bucket_key(self, remote_path: ParseResult) -> Tuple[str, str]:
        assert remote_path.hostname is not None
        return remote_path.hostname, remote_path.path.lstrip("/")

    def upload_fileobj(
        self,
        fileobj: IO[bytes],
        local_path: str,
        *,
        remote_uri: Optional[ParseResult] = None,
        remote_path_prefix: Optional[str] = None,
        remote_path_suffix: Optional[str] = None,
    ) -> ParseResult:
        sha256sum = _hash_fileobj(fileobj)
        remote_uri = _gen_remote_uri(
            fileobj, remote_uri, remote_path_prefix, remote_path_suffix, sha256sum
        )

        import botocore

        bucket, key = self._uri_to_bucket_key(remote_uri)

        try:
            head_response = self.s3_client.head_object(Bucket=bucket, Key=key)
            if head_response["Metadata"]["sha256sum"] == sha256sum:
                # File already exists
                return remote_uri
            else:
                print(f"Uploading {local_path} to {remote_uri}")
                self.s3_client.upload_fileobj(
                    fileobj,
                    bucket,
                    key,
                    ExtraArgs={"Metadata": {"sha256sum": sha256sum}},
                )
                return remote_uri
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] != "404":
                raise

            self.s3_client.upload_fileobj(
                fileobj, bucket, key, ExtraArgs={"Metadata": {"sha256sum": sha256sum}},
            )
            return remote_uri


class AzureBlobClient(AbstractStagingClient):
    """
       Implementation of AbstractStagingClient for Azure Blob storage
    """

    def __init__(self, account_name: str, account_access_key: str):
        try:
            from azure.storage.blob import BlobServiceClient
        except ImportError:
            raise ImportError(
                "Install package azure-storage-blob for azure blob staging support"
                "run ```pip install azure-storage-blob```"
            )
        self.account_name = account_name
        account_url = f"https://{account_name}.blob.core.windows.net"
        self.blob_service_client = BlobServiceClient(
            account_url=account_url, credential=account_access_key
        )

    def download_file(self, uri: ParseResult) -> IO[bytes]:
        """
        Downloads a file from Azure blob storage and returns a TemporaryFile object

        Args:
            uri (urllib.parse.ParseResult): Parsed uri of the file ex: urlparse("wasbs://bucket@account_name.blob.core.windows.net/file.avro")

        Returns:
             TemporaryFile object
        """
        bucket, path = self._uri_to_bucket_key(uri)
        container_client = self.blob_service_client.get_container_client(bucket)
        return container_client.download_blob(path).readall()

    def list_files(self, uri: ParseResult) -> List[str]:
        """
        Lists all the files under a directory in azure blob storage if path has wildcard(*) character.

        Args:
            uri (urllib.parse.ParseResult): Parsed uri of this location

        Returns:
            List[str]: A list containing the full path to the file(s) in the
                    remote staging location.
        """

        bucket, path = self._uri_to_bucket_key(uri)
        if "*" in path:
            regex = re.compile(path.replace("*", ".*?").strip("/"))
            container_client = self.blob_service_client.get_container_client(bucket)
            blob_list = container_client.list_blobs(
                name_starts_with=path.strip("/").split("*")[0]
            )
            # File path should not be in path (file path must be longer than path)
            return [
                f"wasbs://{bucket}@{self.account_name}.blob.core.windows.net/{file}"
                for file in [x.name for x in blob_list]
                if re.match(regex, file) and file not in path
            ]
        else:
            return [
                f"wasbs://{bucket}@{self.account_name}.blob.core.windows.net/{path}"
            ]

    def _uri_to_bucket_key(self, uri: ParseResult) -> Tuple[str, str]:
        assert uri.hostname == f"{self.account_name}.blob.core.windows.net"
        assert uri.username
        bucket = uri.username
        key = uri.path.lstrip("/")
        return bucket, key

    def upload_fileobj(
        self,
        fileobj: IO[bytes],
        local_path: str,
        *,
        remote_uri: Optional[ParseResult] = None,
        remote_path_prefix: Optional[str] = None,
        remote_path_suffix: Optional[str] = None,
    ) -> ParseResult:
        remote_uri = _gen_remote_uri(
            fileobj, remote_uri, remote_path_prefix, remote_path_suffix, None
        )
        bucket, key = self._uri_to_bucket_key(remote_uri)
        container_client = self.blob_service_client.get_container_client(bucket)
        container_client.upload_blob(name=key, data=fileobj, overwrite=True)
        return remote_uri


class LocalFSClient(AbstractStagingClient):
    """
       Implementation of AbstractStagingClient for local file
       Note: The is used for E2E tests.
    """

    def __init__(self):
        pass

    def download_file(self, uri: ParseResult) -> IO[bytes]:
        """
        Reads a local file from the disk

        Args:
            uri (urllib.parse.ParseResult): Parsed uri of the file ex: urlparse("file:///folder/file.avro")
        Returns:
            TemporaryFile object
        """
        url = uri.path
        file_obj = open(url, "rb")
        return file_obj

    def list_files(self, uri: ParseResult) -> List[str]:
        raise NotImplementedError("list files not implemented for Local file")

    def _uri_to_path(self, uri: ParseResult) -> str:
        return uri.path

    def upload_fileobj(
        self,
        fileobj: IO[bytes],
        local_path: str,
        *,
        remote_uri: Optional[ParseResult] = None,
        remote_path_prefix: Optional[str] = None,
        remote_path_suffix: Optional[str] = None,
    ) -> ParseResult:

        remote_uri = _gen_remote_uri(
            fileobj, remote_uri, remote_path_prefix, remote_path_suffix, None
        )
        remote_file_path = self._uri_to_path(remote_uri)
        os.makedirs(os.path.dirname(remote_file_path), exist_ok=True)
        with open(remote_file_path, "wb") as fdest:
            shutil.copyfileobj(fileobj, fdest)
        return remote_uri


def _s3_client(config: Config = None):
    if config is None:
        endpoint_url = None
    else:
        endpoint_url = config.get(opt.S3_ENDPOINT_URL, None)
    return S3Client(endpoint_url=endpoint_url)


def _s3a_client(config: Config = None):
    if config is None:
        endpoint_url = None
    else:
        endpoint_url = config.get(opt.S3_ENDPOINT_URL, None)
    return S3Client(endpoint_url=endpoint_url, url_scheme="s3a")


def _gcs_client(config: Config = None):
    return GCSClient()


def _azure_blob_client(config: Config = None):
    if config is None:
        raise Exception("Azure blob client requires config")
    account_name = config.get(opt.AZURE_BLOB_ACCOUNT_NAME, None)
    account_access_key = config.get(opt.AZURE_BLOB_ACCOUNT_ACCESS_KEY, None)
    if account_name is None or account_access_key is None:
        raise Exception(
            f"Azure blob client requires {opt.AZURE_BLOB_ACCOUNT_NAME} and {opt.AZURE_BLOB_ACCOUNT_ACCESS_KEY} set in config"
        )
    return AzureBlobClient(account_name, account_access_key)


def _local_fs_client(config: Config = None):
    return LocalFSClient()


storage_clients = {
    GS: _gcs_client,
    S3: _s3_client,
    S3A: _s3a_client,
    AZURE_SCHEME: _azure_blob_client,
    LOCAL_FILE: _local_fs_client,
}


def get_staging_client(scheme, config: Config = None) -> AbstractStagingClient:
    """
    Initialization of a specific client object(GCSClient, S3Client etc.)

    Args:
        scheme (str): uri scheme: s3, gs or file
        config (Config): additional configuration

    Returns:
        An object of concrete implementation of AbstractStagingClient
    """
    try:
        return storage_clients[scheme](config)
    except ValueError:
        raise Exception(
            f"Could not identify file scheme {scheme}. Only gs://, file://, s3:// and wasbs:// (for Azure) are supported"
        )



# SPDX-License-Identifier: Apache-2.0
# Copyright 2018-2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import time
from http import HTTPStatus

import grpc
import requests
from google.auth.exceptions import DefaultCredentialsError
from google.auth.transport import requests as grequests

from feast.config import Config
from feast.constants import AuthProvider
from feast.constants import ConfigOptions as opt


def get_auth_metadata_plugin(config: Config) -> grpc.AuthMetadataPlugin:
    """
    Get an Authentication Metadata Plugin. This plugin is used in gRPC to
    sign requests. Please see the following URL for more details
    https://grpc.github.io/grpc/python/_modules/grpc.html#AuthMetadataPlugin

    New plugins can be added to this function. For the time being we only
    support Google Open ID authentication.

    Returns: Returns an implementation of grpc.AuthMetadataPlugin

    Args:
        config: Feast Configuration object
    """
    if AuthProvider(config.get(opt.AUTH_PROVIDER)) == AuthProvider.GOOGLE:
        return GoogleOpenIDAuthMetadataPlugin(config)
    elif AuthProvider(config.get(opt.AUTH_PROVIDER)) == AuthProvider.OAUTH:
        return OAuthMetadataPlugin(config)
    else:
        raise RuntimeError(
            "Could not determine OAuth provider."
            'Must be set to either "google" or "oauth"'
        )


class OAuthMetadataPlugin(grpc.AuthMetadataPlugin):
    """A `gRPC AuthMetadataPlugin`_ that inserts the credentials into each
    request.

    .. _gRPC AuthMetadataPlugin:
        http://www.grpc.io/grpc/python/grpc.html#grpc.AuthMetadataPlugin
    """

    def __init__(self, config: Config):
        """
        Initializes an OAuthMetadataPlugin, used to sign gRPC requests
        Args:
            config: Feast Configuration object
        """
        super(OAuthMetadataPlugin, self).__init__()

        self._static_token = None
        self._token = None

        # If provided, set a static token
        if config.exists(opt.AUTH_TOKEN):
            self._static_token = config.get(opt.AUTH_TOKEN)
            self._refresh_token(config)
        elif (
            config.exists(opt.OAUTH_GRANT_TYPE)
            and config.exists(opt.OAUTH_CLIENT_ID)
            and config.exists(opt.OAUTH_CLIENT_SECRET)
            and config.exists(opt.OAUTH_AUDIENCE)
            and config.exists(opt.OAUTH_TOKEN_REQUEST_URL)
        ):
            self._refresh_token(config)
        else:
            raise RuntimeError(
                " Please ensure that the "
                "necessary parameters are passed to the client - "
                "oauth_grant_type, oauth_client_id, oauth_client_secret, "
                "oauth_audience, oauth_token_request_url."
            )

    def get_signed_meta(self):
        """ Creates a signed authorization metadata token."""
        return (("authorization", "Bearer {}".format(self._token)),)

    def _refresh_token(self, config: Config):
        """ Refreshes OAuth token and persists it in memory """

        # Use static token if available
        if self._static_token:
            self._token = self._static_token
            return

        headers_token = {"content-type": "application/json"}
        data_token_dict = {
            "grant_type": config.get(opt.OAUTH_GRANT_TYPE),
            "client_id": config.get(opt.OAUTH_CLIENT_ID),
            "client_secret": config.get(opt.OAUTH_CLIENT_SECRET),
            "audience": config.get(opt.OAUTH_AUDIENCE),
        }
        data_token = json.dumps(data_token_dict)
        response_token = requests.post(
            config.get(opt.OAUTH_TOKEN_REQUEST_URL),
            headers=headers_token,
            data=data_token,
        )
        if response_token.status_code == HTTPStatus.OK:
            self._token = response_token.json().get("access_token")
        else:
            raise RuntimeError(
                f"Could not fetch OAuth token, got response : {response_token.status_code}"
            )

    def set_static_token(self, token):
        """
        Define a static token to return

        Args:
            token: String token
        """
        self._static_token = token

    def __call__(self, context, callback):
        """Passes authorization metadata into the given callback.

        Args:
            context (grpc.AuthMetadataContext): The RPC context.
            callback (grpc.AuthMetadataPluginCallback): The callback that will
                be invoked to pass in the authorization metadata.
        """
        callback(self.get_signed_meta(), None)


class GoogleOpenIDAuthMetadataPlugin(grpc.AuthMetadataPlugin):
    """A `gRPC AuthMetadataPlugin`_ that inserts the credentials into each
    request.

    .. _gRPC AuthMetadataPlugin:
        http://www.grpc.io/grpc/python/grpc.html#grpc.AuthMetadataPlugin
    """

    def __init__(self, config: Config):
        """
        Initializes a GoogleOpenIDAuthMetadataPlugin, used to sign gRPC requests
        Args:
            config: Feast Configuration object
        """
        super(GoogleOpenIDAuthMetadataPlugin, self).__init__()

        self._static_token = None
        self._token = None
        self._token_expiry_ts = time.time()

        # If provided, set a static token
        if config.exists(opt.AUTH_TOKEN):
            self._static_token = config.get(opt.AUTH_TOKEN)

        self._request = RequestWithTimeout(timeout=5)
        self._refresh_token()

    def get_signed_meta(self):
        """ Creates a signed authorization metadata token."""

        if time.time() > self._token_expiry_ts:
            self._refresh_token()
        return (("authorization", "Bearer {}".format(self._token)),)

    def _refresh_token(self):
        """ Refreshes Google ID token and persists it in memory """

        # Use static token if available
        if self._static_token:
            self._token = self._static_token
            return

        from google.oauth2.id_token import fetch_id_token, verify_oauth2_token

        try:
            self._token = fetch_id_token(self._request, audience="feast.dev")
            self._token_expiry_ts = verify_oauth2_token(self._token, self._request)[
                "exp"
            ]
            return
        except DefaultCredentialsError:
            pass

        # Try to use Google Auth library to find ID Token
        from google import auth as google_auth

        try:
            credentials, _ = google_auth.default(["openid", "email"])
            credentials.refresh(self._request)
            if hasattr(credentials, "id_token"):
                self._token = credentials.id_token
                self._token_expiry_ts = verify_oauth2_token(self._token, self._request)[
                    "exp"
                ]
                return
        except DefaultCredentialsError:
            pass  # Could not determine credentials, skip

        # Raise exception otherwise
        raise RuntimeError(
            "Could not determine Google ID token. Ensure that a service account can be found by setting"
            " the GOOGLE_APPLICATION_CREDENTIALS environmental variable to its path."
        )

    def set_static_token(self, token):
        """
        Define a static token to return

        Args:
            token: String token
        """
        self._static_token = token

    def __call__(self, context, callback):
        """Passes authorization metadata into the given callback.

        Args:
            context (grpc.AuthMetadataContext): The RPC context.
            callback (grpc.AuthMetadataPluginCallback): The callback that will
                be invoked to pass in the authorization metadata.
        """
        callback(self.get_signed_meta(), None)


class RequestWithTimeout(grequests.Request):
    def __init__(self, *args, timeout=None, **kwargs):
        self._timeout = timeout
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        timeout = kwargs.pop("timeout", self._timeout)
        return super().__call__(*args, timeout=timeout, **kwargs)



import logging
import time
from concurrent import futures

import grpc
from google.protobuf.timestamp_pb2 import Timestamp

from feast.protos.feast.core import CoreService_pb2_grpc as Core
from feast.protos.feast.core.CoreService_pb2 import (
    ApplyEntityRequest,
    ApplyEntityResponse,
    ApplyFeatureTableRequest,
    ApplyFeatureTableResponse,
    DeleteFeatureTableRequest,
    DeleteFeatureTableResponse,
    GetEntityRequest,
    GetEntityResponse,
    GetFeastCoreVersionResponse,
    GetFeatureTableRequest,
    GetFeatureTableResponse,
    ListEntitiesRequest,
    ListEntitiesResponse,
    ListFeatureTablesRequest,
    ListFeatureTablesResponse,
    ListProjectsResponse,
)
from feast.protos.feast.core.Entity_pb2 import Entity as EntityProto
from feast.protos.feast.core.Entity_pb2 import EntityMeta
from feast.protos.feast.core.FeatureTable_pb2 import FeatureTable as FeatureTableProto
from feast.protos.feast.core.FeatureTable_pb2 import FeatureTableMeta

_logger = logging.getLogger(__name__)

_ONE_DAY_IN_SECONDS = 60 * 60 * 24
_SIGNATURE_HEADER_KEY = "authorization"


class DisallowAuthInterceptor(grpc.ServerInterceptor):
    def __init__(self):
        def abort(ignored_request, context):
            context.abort(grpc.StatusCode.UNAUTHENTICATED, "Invalid signature")

        self._abortion = grpc.unary_unary_rpc_method_handler(abort)

    def intercept_service(self, continuation, handler_call_details):
        print(handler_call_details.invocation_metadata)
        if "Bearer" in handler_call_details.invocation_metadata[0][1]:
            return self._abortion
        else:
            return continuation(handler_call_details)


class AllowAuthInterceptor(grpc.ServerInterceptor):
    def __init__(self):
        def abort(ignored_request, context):
            context.abort(grpc.StatusCode.UNAUTHENTICATED, "Invalid signature")

        self._abortion = grpc.unary_unary_rpc_method_handler(abort)

    def intercept_service(self, continuation, handler_call_details):
        print(handler_call_details.invocation_metadata)
        if "Bearer" in handler_call_details.invocation_metadata[0][1]:
            return continuation(handler_call_details)
        else:
            return self._abortion


class CoreServicer(Core.CoreServiceServicer):
    def __init__(self):
        self._feature_tables = dict()
        self._entities = dict()
        self._projects = ["default"]

    def GetFeastCoreVersion(self, request, context):
        return GetFeastCoreVersionResponse(version="0.10.0")

    def GetFeatureTable(self, request: GetFeatureTableRequest, context):
        filtered_table = [
            table
            for table in self._feature_tables.values()
            if table.spec.name == request.name
        ]
        return GetFeatureTableResponse(table=filtered_table[0])

    def ListFeatureTables(self, request: ListFeatureTablesRequest, context):

        filtered_feature_table_response = list(self._feature_tables.values())

        return ListFeatureTablesResponse(tables=filtered_feature_table_response)

    def ApplyFeatureTable(self, request: ApplyFeatureTableRequest, context):
        feature_table_spec = request.table_spec

        feature_table_meta = FeatureTableMeta(created_timestamp=Timestamp(seconds=10),)
        applied_feature_table = FeatureTableProto(
            spec=feature_table_spec, meta=feature_table_meta
        )
        self._feature_tables[feature_table_spec.name] = applied_feature_table

        _logger.info(
            "registered feature table "
            + feature_table_spec.name
            + " with "
            + str(len(feature_table_spec.entities))
            + " entities and "
            + str(len(feature_table_spec.features))
            + " features"
        )

        return ApplyFeatureTableResponse(table=applied_feature_table,)

    def DeleteFeatureTable(self, request: DeleteFeatureTableRequest, context):
        del self._feature_tables[request.name]
        return DeleteFeatureTableResponse()

    def GetEntity(self, request: GetEntityRequest, context):
        filtered_entities = [
            entity
            for entity in self._entities.values()
            if entity.spec.name == request.name
        ]
        return GetEntityResponse(entity=filtered_entities[0])

    def ListEntities(self, request: ListEntitiesRequest, context):

        filtered_entities_response = list(self._entities.values())

        return ListEntitiesResponse(entities=filtered_entities_response)

    def ListProjects(self, request, context):
        return ListProjectsResponse(projects=self._projects)

    def ApplyEntity(self, request: ApplyEntityRequest, context):
        entity_spec = request.spec

        entity_meta = EntityMeta(created_timestamp=Timestamp(seconds=10),)
        applied_entity = EntityProto(spec=entity_spec, meta=entity_meta)
        self._entities[entity_spec.name] = applied_entity

        _logger.info(
            "registered entity "
            + entity_spec.name
            + " with "
            + str(entity_spec.value_type)
            + " value"
        )

        return ApplyEntityResponse(entity=applied_entity,)


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    Core.add_CoreServiceServicer_to_server(CoreServicer(), server)
    server.add_insecure_port("[::]:50051")
    server.start()
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


if __name__ == "__main__":
    logging.basicConfig()
    serve()



import struct
from typing import List, Tuple

from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.protos.feast.types.Value_pb2 import ValueType


def _serialize_val(value_type, v: ValueProto) -> Tuple[bytes, int]:
    if value_type == "string_val":
        return v.string_val.encode("utf8"), ValueType.STRING
    elif value_type == "bytes_val":
        return v.bytes_val, ValueType.BYTES
    elif value_type == "int32_val":
        return struct.pack("<i", v.int32_val), ValueType.INT32
    elif value_type == "int64_val":
        return struct.pack("<l", v.int64_val), ValueType.INT64
    else:
        raise ValueError(f"Value type not supported for Firestore: {v}")


def serialize_entity_key(entity_key: EntityKeyProto) -> bytes:
    """
    Serialize entity key to a bytestring so it can be used as a lookup key in a hash table.

    We need this encoding to be stable; therefore we cannot just use protobuf serialization
    here since it does not guarantee that two proto messages containing the same data will
    serialize to the same byte string[1].

    [1] https://developers.google.com/protocol-buffers/docs/encoding
    """
    sorted_keys, sorted_values = zip(
        *sorted(zip(entity_key.join_keys, entity_key.entity_values))
    )

    output: List[bytes] = []
    for k in sorted_keys:
        output.append(struct.pack("<I", ValueType.STRING))
        output.append(k.encode("utf8"))
    for v in sorted_values:
        val_bytes, value_type = _serialize_val(v.WhichOneof("val"), v)

        output.append(struct.pack("<I", value_type))

        output.append(struct.pack("<I", len(val_bytes)))
        output.append(val_bytes)

    return b"".join(output)



# Copyright 2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
from concurrent import futures
from contextlib import closing

import grpc
import pytest

from feast.client import Client
from feast.entity import Entity
from feast.protos.feast.core import CoreService_pb2_grpc as Core
from feast.value_type import ValueType
from feast_core_server import CoreServicer


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


free_port = find_free_port()


class TestEntity:
    @pytest.fixture(scope="function")
    def server(self):
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        Core.add_CoreServiceServicer_to_server(CoreServicer(), server)
        server.add_insecure_port(f"[::]:{free_port}")
        server.start()
        yield server
        server.stop(0)

    @pytest.fixture
    def client(self, server):
        return Client(core_url=f"localhost:{free_port}")

    def test_entity_import_export_yaml(self):

        test_entity = Entity(
            name="car_driver_entity",
            description="Driver entity for car rides",
            value_type=ValueType.STRING,
            labels={"team": "matchmaking"},
        )

        # Create a string YAML representation of the entity
        string_yaml = test_entity.to_yaml()

        # Create a new entity object from the YAML string
        actual_entity_from_string = Entity.from_yaml(string_yaml)

        # Ensure equality is upheld to original entity
        assert test_entity == actual_entity_from_string


def test_join_key_default():
    entity = Entity("my-entity", description="My entity", value_type=ValueType.STRING)
    assert entity.join_key == "my-entity"


def test_entity_class_contains_labels():
    entity = Entity(
        "my-entity",
        description="My entity",
        value_type=ValueType.STRING,
        labels={"key1": "val1", "key2": "val2"},
    )
    assert "key1" in entity.labels.keys() and entity.labels["key1"] == "val1"
    assert "key2" in entity.labels.keys() and entity.labels["key2"] == "val2"


def test_entity_without_labels_empty_dict():
    entity = Entity("my-entity", description="My entity", value_type=ValueType.STRING)
    assert entity.labels == dict()
    assert len(entity.labels) == 0


def test_entity_without_description():
    Entity("my-entity", value_type=ValueType.STRING)



from datetime import datetime
from typing import Callable, List, Optional, Union

import pandas as pd
import pyarrow
import pytz

from feast.data_source import DataSource, FileSource
from feast.feature_view import FeatureView
from feast.infra.offline_stores.offline_store import OfflineStore, RetrievalJob
from feast.infra.provider import (
    ENTITY_DF_EVENT_TIMESTAMP_COL,
    _get_requested_feature_views_to_features_dict,
    _run_field_mapping,
)
from feast.registry import Registry
from feast.repo_config import RepoConfig


class FileRetrievalJob(RetrievalJob):
    def __init__(self, evaluation_function: Callable):
        """Initialize a lazy historical retrieval job"""

        # The evaluation function executes a stored procedure to compute a historical retrieval.
        self.evaluation_function = evaluation_function

    def to_df(self):
        # Only execute the evaluation function to build the final historical retrieval dataframe at the last moment.
        df = self.evaluation_function()
        return df


class FileOfflineStore(OfflineStore):
    @staticmethod
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: Registry,
        project: str,
    ) -> FileRetrievalJob:
        if not isinstance(entity_df, pd.DataFrame):
            raise ValueError(
                f"Please provide an entity_df of type {type(pd.DataFrame)} instead of type {type(entity_df)}"
            )
        if ENTITY_DF_EVENT_TIMESTAMP_COL not in entity_df.columns:
            raise ValueError(
                f"Please provide an entity_df with a column named {ENTITY_DF_EVENT_TIMESTAMP_COL} representing the time of events."
            )

        feature_views_to_features = _get_requested_feature_views_to_features_dict(
            feature_refs, feature_views
        )

        # Create lazy function that is only called from the RetrievalJob object
        def evaluate_historical_retrieval():

            # Make sure all event timestamp fields are tz-aware. We default tz-naive fields to UTC
            entity_df[ENTITY_DF_EVENT_TIMESTAMP_COL] = entity_df[
                ENTITY_DF_EVENT_TIMESTAMP_COL
            ].apply(lambda x: x if x.tzinfo is not None else x.replace(tzinfo=pytz.utc))

            # Create a copy of entity_df to prevent modifying the original
            entity_df_with_features = entity_df.copy()

            # Convert event timestamp column to datetime and normalize time zone to UTC
            # This is necessary to avoid issues with pd.merge_asof
            entity_df_with_features[ENTITY_DF_EVENT_TIMESTAMP_COL] = pd.to_datetime(
                entity_df_with_features[ENTITY_DF_EVENT_TIMESTAMP_COL], utc=True
            )

            # Sort event timestamp values
            entity_df_with_features = entity_df_with_features.sort_values(
                ENTITY_DF_EVENT_TIMESTAMP_COL
            )

            # Load feature view data from sources and join them incrementally
            for feature_view, features in feature_views_to_features.items():
                event_timestamp_column = feature_view.input.event_timestamp_column
                created_timestamp_column = feature_view.input.created_timestamp_column

                # Read offline parquet data in pyarrow format
                table = pyarrow.parquet.read_table(feature_view.input.path)

                # Rename columns by the field mapping dictionary if it exists
                if feature_view.input.field_mapping is not None:
                    table = _run_field_mapping(table, feature_view.input.field_mapping)

                # Convert pyarrow table to pandas dataframe
                df_to_join = table.to_pandas()

                # Make sure all timestamp fields are tz-aware. We default tz-naive fields to UTC
                df_to_join[event_timestamp_column] = df_to_join[
                    event_timestamp_column
                ].apply(
                    lambda x: x if x.tzinfo is not None else x.replace(tzinfo=pytz.utc)
                )
                if created_timestamp_column:
                    df_to_join[created_timestamp_column] = df_to_join[
                        created_timestamp_column
                    ].apply(
                        lambda x: x
                        if x.tzinfo is not None
                        else x.replace(tzinfo=pytz.utc)
                    )

                # Sort dataframe by the event timestamp column
                df_to_join = df_to_join.sort_values(event_timestamp_column)

                # Build a list of all the features we should select from this source
                feature_names = []
                for feature in features:
                    # Modify the separator for feature refs in column names to double underscore. We are using
                    # double underscore as separator for consistency with other databases like BigQuery,
                    # where there are very few characters available for use as separators
                    prefixed_feature_name = f"{feature_view.name}__{feature}"

                    # Add the feature name to the list of columns
                    feature_names.append(prefixed_feature_name)

                    # Ensure that the source dataframe feature column includes the feature view name as a prefix
                    df_to_join.rename(
                        columns={feature: prefixed_feature_name}, inplace=True,
                    )

                # Build a list of entity columns to join on (from the right table)
                join_keys = []
                for entity_name in feature_view.entities:
                    entity = registry.get_entity(entity_name, project)
                    join_keys.append(entity.join_key)
                right_entity_columns = join_keys
                right_entity_key_columns = [
                    event_timestamp_column
                ] + right_entity_columns

                # Remove all duplicate entity keys (using created timestamp)
                right_entity_key_sort_columns = right_entity_key_columns
                if created_timestamp_column:
                    # If created_timestamp is available, use it to dedupe deterministically
                    right_entity_key_sort_columns = right_entity_key_sort_columns + [
                        created_timestamp_column
                    ]

                df_to_join.sort_values(by=right_entity_key_sort_columns, inplace=True)
                df_to_join = df_to_join.groupby(by=right_entity_key_columns).last()
                df_to_join.reset_index(inplace=True)

                # Select only the columns we need to join from the feature dataframe
                df_to_join = df_to_join[right_entity_key_columns + feature_names]

                # Do point in-time-join between entity_df and feature dataframe
                entity_df_with_features = pd.merge_asof(
                    entity_df_with_features,
                    df_to_join,
                    left_on=ENTITY_DF_EVENT_TIMESTAMP_COL,
                    right_on=event_timestamp_column,
                    by=right_entity_columns,
                    tolerance=feature_view.ttl,
                )

                # Remove right (feature table/view) event_timestamp column.
                if event_timestamp_column != ENTITY_DF_EVENT_TIMESTAMP_COL:
                    entity_df_with_features.drop(
                        columns=[event_timestamp_column], inplace=True
                    )

                # Ensure that we delete dataframes to free up memory
                del df_to_join

            # Move "datetime" column to front
            current_cols = entity_df_with_features.columns.tolist()
            current_cols.remove(ENTITY_DF_EVENT_TIMESTAMP_COL)
            entity_df_with_features = entity_df_with_features[
                [ENTITY_DF_EVENT_TIMESTAMP_COL] + current_cols
            ]

            return entity_df_with_features

        job = FileRetrievalJob(evaluation_function=evaluate_historical_retrieval)
        return job

    @staticmethod
    def pull_latest_from_table_or_query(
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        event_timestamp_column: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> pyarrow.Table:
        assert isinstance(data_source, FileSource)

        source_df = pd.read_parquet(data_source.path)
        # Make sure all timestamp fields are tz-aware. We default tz-naive fields to UTC
        source_df[event_timestamp_column] = source_df[event_timestamp_column].apply(
            lambda x: x if x.tzinfo is not None else x.replace(tzinfo=pytz.utc)
        )
        if created_timestamp_column:
            source_df[created_timestamp_column] = source_df[
                created_timestamp_column
            ].apply(lambda x: x if x.tzinfo is not None else x.replace(tzinfo=pytz.utc))

        ts_columns = (
            [event_timestamp_column, created_timestamp_column]
            if created_timestamp_column
            else [event_timestamp_column]
        )

        source_df.sort_values(by=ts_columns, inplace=True)

        filtered_df = source_df[
            (source_df[event_timestamp_column] >= start_date)
            & (source_df[event_timestamp_column] < end_date)
        ]
        last_values_df = filtered_df.groupby(by=join_key_columns).last()

        # make driver_id a normal column again
        last_values_df.reset_index(inplace=True)

        table = pyarrow.Table.from_pandas(
            last_values_df[join_key_columns + feature_name_columns + ts_columns]
        )

        return table



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timezone
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
import pyarrow as pa
from google.protobuf.json_format import MessageToDict
from google.protobuf.timestamp_pb2 import Timestamp
from pyarrow.lib import TimestampType

from feast.protos.feast.types.Value_pb2 import (
    BoolList,
    BytesList,
    DoubleList,
    FloatList,
    Int32List,
    Int64List,
    StringList,
)
from feast.protos.feast.types.Value_pb2 import Value as ProtoValue
from feast.protos.feast.types.Value_pb2 import ValueType as ProtoValueType
from feast.value_type import ValueType


def feast_value_type_to_python_type(field_value_proto: ProtoValue) -> Any:
    """
    Converts field value Proto to Dict and returns each field's Feast Value Type value
    in their respective Python value.

    Args:
        field_value_proto: Field value Proto

    Returns:
        Python native type representation/version of the given field_value_proto
    """
    field_value_dict = MessageToDict(field_value_proto)

    for k, v in field_value_dict.items():
        if k == "int64Val":
            return int(v)
        if k == "bytesVal":
            return bytes(v)
        if (k == "int64ListVal") or (k == "int32ListVal"):
            return [int(item) for item in v["val"]]
        if (k == "floatListVal") or (k == "doubleListVal"):
            return [float(item) for item in v["val"]]
        if k == "stringListVal":
            return [str(item) for item in v["val"]]
        if k == "bytesListVal":
            return [bytes(item) for item in v["val"]]
        if k == "boolListVal":
            return [bool(item) for item in v["val"]]

        if k in ["int32Val", "floatVal", "doubleVal", "stringVal", "boolVal"]:
            return v
        else:
            raise TypeError(
                f"Casting to Python native type for type {k} failed. "
                f"Type {k} not found"
            )


def python_type_to_feast_value_type(
    name: str, value, recurse: bool = True
) -> ValueType:
    """
    Finds the equivalent Feast Value Type for a Python value. Both native
    and Pandas types are supported. This function will recursively look
    for nested types when arrays are detected. All types must be homogenous.

    Args:
        name: Name of the value or field
        value: Value that will be inspected
        recurse: Whether to recursively look for nested types in arrays

    Returns:
        Feast Value Type
    """

    type_name = type(value).__name__

    type_map = {
        "int": ValueType.INT64,
        "str": ValueType.STRING,
        "float": ValueType.DOUBLE,
        "bytes": ValueType.BYTES,
        "float64": ValueType.DOUBLE,
        "float32": ValueType.FLOAT,
        "int64": ValueType.INT64,
        "uint64": ValueType.INT64,
        "int32": ValueType.INT32,
        "uint32": ValueType.INT32,
        "uint8": ValueType.INT32,
        "int8": ValueType.INT32,
        "bool": ValueType.BOOL,
        "timedelta": ValueType.INT64,
        "datetime64[ns]": ValueType.INT64,
        "datetime64[ns, tz]": ValueType.INT64,
        "category": ValueType.STRING,
    }

    if type_name in type_map:
        return type_map[type_name]

    if type_name == "ndarray" or isinstance(value, list):
        if recurse:

            # Convert to list type
            list_items = pd.core.series.Series(value)

            # This is the final type which we infer from the list
            common_item_value_type = None
            for item in list_items:
                if isinstance(item, ProtoValue):
                    current_item_value_type = _proto_str_to_value_type(
                        item.WhichOneof("val")
                    )
                else:
                    # Get the type from the current item, only one level deep
                    current_item_value_type = python_type_to_feast_value_type(
                        name=name, value=item, recurse=False
                    )
                # Validate whether the type stays consistent
                if (
                    common_item_value_type
                    and not common_item_value_type == current_item_value_type
                ):
                    raise ValueError(
                        f"List value type for field {name} is inconsistent. "
                        f"{common_item_value_type} different from "
                        f"{current_item_value_type}."
                    )
                common_item_value_type = current_item_value_type
            if common_item_value_type is None:
                raise ValueError(
                    f"field {name} cannot have null values for type inference."
                )
            return ValueType[common_item_value_type.name + "_LIST"]
        else:
            raise ValueError(
                f"Value type for field {name} is {value.dtype.__str__()} but "
                f"recursion is not allowed. Array types can only be one level "
                f"deep."
            )

    return type_map[value.dtype.__str__()]


def _pd_datetime_to_timestamp_proto(dtype, value) -> Timestamp:
    """
    Converts a Pandas datetime to a Timestamp Proto

    Args:
        dtype: Pandas datatype
        value: Value of datetime

    Returns:
        Timestamp protobuf value
    """

    if type(value) in [np.float64, np.float32, np.int32, np.int64]:
        return Timestamp(seconds=int(value))
    if dtype.__str__() == "datetime64[ns]":
        # If timestamp does not contain timezone, we assume it is of local
        # timezone and adjust it to UTC
        local_timezone = datetime.now(timezone.utc).astimezone().tzinfo
        value = value.tz_localize(local_timezone).tz_convert("UTC").tz_localize(None)
        return Timestamp(seconds=int(value.timestamp()))
    if dtype.__str__() == "datetime64[ns, UTC]":
        return Timestamp(seconds=int(value.timestamp()))
    else:
        return Timestamp(seconds=np.datetime64(value).astype("int64") // 1000000)  # type: ignore


def _type_err(item, dtype):
    raise ValueError(f'Value "{item}" is of type {type(item)} not of type {dtype}')


def _python_value_to_proto_value(feast_value_type, value) -> ProtoValue:
    """
    Converts a Python (native, pandas) value to a Feast Proto Value based
    on a provided value type

    Args:
        feast_value_type: The target value type
        value: Value that will be converted

    Returns:
        Feast Value Proto
    """

    # Detect list type and handle separately
    if "list" in feast_value_type.name.lower():

        if feast_value_type == ValueType.FLOAT_LIST:
            return ProtoValue(
                float_list_val=FloatList(
                    val=[
                        item
                        if type(item) in [np.float32, np.float64]
                        else _type_err(item, np.float32)
                        for item in value
                    ]
                )
            )

        if feast_value_type == ValueType.DOUBLE_LIST:
            return ProtoValue(
                double_list_val=DoubleList(
                    val=[
                        item
                        if type(item) in [np.float64, np.float32]
                        else _type_err(item, np.float64)
                        for item in value
                    ]
                )
            )

        if feast_value_type == ValueType.INT32_LIST:
            return ProtoValue(
                int32_list_val=Int32List(
                    val=[
                        item if type(item) is np.int32 else _type_err(item, np.int32)
                        for item in value
                    ]
                )
            )

        if feast_value_type == ValueType.INT64_LIST:
            return ProtoValue(
                int64_list_val=Int64List(
                    val=[
                        item
                        if type(item) in [np.int64, np.int32]
                        else _type_err(item, np.int64)
                        for item in value
                    ]
                )
            )

        if feast_value_type == ValueType.STRING_LIST:
            return ProtoValue(
                string_list_val=StringList(
                    val=[
                        item
                        if type(item) in [np.str_, str]
                        else _type_err(item, np.str_)
                        for item in value
                    ]
                )
            )

        if feast_value_type == ValueType.BOOL_LIST:
            return ProtoValue(
                bool_list_val=BoolList(
                    val=[
                        item
                        if type(item) in [np.bool_, bool]
                        else _type_err(item, np.bool_)
                        for item in value
                    ]
                )
            )

        if feast_value_type == ValueType.BYTES_LIST:
            return ProtoValue(
                bytes_list_val=BytesList(
                    val=[
                        item
                        if type(item) in [np.bytes_, bytes]
                        else _type_err(item, np.bytes_)
                        for item in value
                    ]
                )
            )

    # Handle scalar types below
    else:
        if pd.isnull(value):
            return ProtoValue()
        elif feast_value_type == ValueType.INT32:
            return ProtoValue(int32_val=int(value))
        elif feast_value_type == ValueType.INT64:
            return ProtoValue(int64_val=int(value))
        elif feast_value_type == ValueType.FLOAT:
            return ProtoValue(float_val=float(value))
        elif feast_value_type == ValueType.DOUBLE:
            assert type(value) is float or np.float64
            return ProtoValue(double_val=value)
        elif feast_value_type == ValueType.STRING:
            return ProtoValue(string_val=str(value))
        elif feast_value_type == ValueType.BYTES:
            assert type(value) is bytes
            return ProtoValue(bytes_val=value)
        elif feast_value_type == ValueType.BOOL:
            assert type(value) is bool
            return ProtoValue(bool_val=value)

    raise Exception(f"Unsupported data type: ${str(type(value))}")


def python_value_to_proto_value(value: Any) -> ProtoValue:
    value_type = python_type_to_feast_value_type("", value)
    return _python_value_to_proto_value(value_type, value)


def _proto_str_to_value_type(proto_str: str) -> ValueType:
    """
    Returns Feast ValueType given Feast ValueType string.

    Args:
        proto_str: str

    Returns:
        A variant of ValueType.
    """
    type_map = {
        "int32_val": ValueType.INT32,
        "int64_val": ValueType.INT64,
        "double_val": ValueType.DOUBLE,
        "float_val": ValueType.FLOAT,
        "string_val": ValueType.STRING,
        "bytes_val": ValueType.BYTES,
        "bool_val": ValueType.BOOL,
        "int32_list_val": ValueType.INT32_LIST,
        "int64_list_val": ValueType.INT64_LIST,
        "double_list_val": ValueType.DOUBLE_LIST,
        "float_list_val": ValueType.FLOAT_LIST,
        "string_list_val": ValueType.STRING_LIST,
        "bytes_list_val": ValueType.BYTES_LIST,
        "bool_list_val": ValueType.BOOL_LIST,
    }

    return type_map[proto_str]


def pa_to_feast_value_attr(pa_type: object):
    """
    Returns the equivalent Feast ValueType string for the given pa.lib type.

    Args:
        pa_type (object):
            PyArrow type.

    Returns:
        str:
            Feast attribute name in Feast ValueType string-ed representation.
    """
    # Mapping of PyArrow type to attribute name in Feast ValueType strings
    type_map = {
        "timestamp[ms]": "int64_val",
        "int32": "int32_val",
        "int64": "int64_val",
        "double": "double_val",
        "float": "float_val",
        "string": "string_val",
        "binary": "bytes_val",
        "bool": "bool_val",
        "list<item: int32>": "int32_list_val",
        "list<item: int64>": "int64_list_val",
        "list<item: double>": "double_list_val",
        "list<item: float>": "float_list_val",
        "list<item: string>": "string_list_val",
        "list<item: binary>": "bytes_list_val",
        "list<item: bool>": "bool_list_val",
    }

    return type_map[pa_type.__str__()]


def pa_to_value_type(pa_type: object):
    """
    Returns the equivalent Feast ValueType for the given pa.lib type.

    Args:
        pa_type (object):
            PyArrow type.

    Returns:
        feast.types.Value_pb2.ValueType:
            Feast ValueType.

    """

    # Mapping of PyArrow to attribute name in Feast ValueType
    type_map = {
        "timestamp[ms]": ProtoValueType.INT64,
        "int32": ProtoValueType.INT32,
        "int64": ProtoValueType.INT64,
        "double": ProtoValueType.DOUBLE,
        "float": ProtoValueType.FLOAT,
        "string": ProtoValueType.STRING,
        "binary": ProtoValueType.BYTES,
        "bool": ProtoValueType.BOOL,
        "list<item: int32>": ProtoValueType.INT32_LIST,
        "list<item: int64>": ProtoValueType.INT64_LIST,
        "list<item: double>": ProtoValueType.DOUBLE_LIST,
        "list<item: float>": ProtoValueType.FLOAT_LIST,
        "list<item: string>": ProtoValueType.STRING_LIST,
        "list<item: binary>": ProtoValueType.BYTES_LIST,
        "list<item: bool>": ProtoValueType.BOOL_LIST,
    }
    return type_map[pa_type.__str__()]


def pa_to_feast_value_type(value: pa.lib.ChunkedArray) -> ValueType:
    type_map = {
        "timestamp[ms]": ValueType.INT64,
        "int32": ValueType.INT32,
        "int64": ValueType.INT64,
        "double": ValueType.DOUBLE,
        "float": ValueType.FLOAT,
        "string": ValueType.STRING,
        "binary": ValueType.BYTES,
        "bool": ValueType.BOOL,
        "list<item: int32>": ValueType.INT32_LIST,
        "list<item: int64>": ValueType.INT64_LIST,
        "list<item: double>": ValueType.DOUBLE_LIST,
        "list<item: float>": ValueType.FLOAT_LIST,
        "list<item: string>": ValueType.STRING_LIST,
        "list<item: binary>": ValueType.BYTES_LIST,
        "list<item: bool>": ValueType.BOOL_LIST,
    }
    return type_map[value.type.__str__()]


def pa_column_to_timestamp_proto_column(column: pa.lib.ChunkedArray) -> List[Timestamp]:
    if not isinstance(column.type, TimestampType):
        raise Exception("Only TimestampType columns are allowed")

    proto_column = []
    for val in column:
        timestamp = Timestamp()
        timestamp.FromMicroseconds(micros=int(val.as_py().timestamp() * 1_000_000))
        proto_column.append(timestamp)
    return proto_column


def pa_column_to_proto_column(
    feast_value_type: ValueType, column: pa.lib.ChunkedArray
) -> List[ProtoValue]:
    type_map: Dict[ValueType, Union[str, Dict[str, Any]]] = {
        ValueType.INT32: "int32_val",
        ValueType.INT64: "int64_val",
        ValueType.FLOAT: "float_val",
        ValueType.DOUBLE: "double_val",
        ValueType.STRING: "string_val",
        ValueType.BYTES: "bytes_val",
        ValueType.BOOL: "bool_val",
        ValueType.BOOL_LIST: {"bool_list_val": BoolList},
        ValueType.BYTES_LIST: {"bytes_list_val": BytesList},
        ValueType.STRING_LIST: {"string_list_val": StringList},
        ValueType.FLOAT_LIST: {"float_list_val": FloatList},
        ValueType.DOUBLE_LIST: {"double_list_val": DoubleList},
        ValueType.INT32_LIST: {"int32_list_val": Int32List},
        ValueType.INT64_LIST: {"int64_list_val": Int64List},
    }

    value: Union[str, Dict[str, Any]] = type_map[feast_value_type]
    # Process list types
    if isinstance(value, dict):
        list_param_name = list(value.keys())[0]
        return [
            ProtoValue(**{list_param_name: value[list_param_name](val=x.as_py())})
            for x in column
        ]
    else:
        return [ProtoValue(**{value: x.as_py()}) for x in column]



# Copyright 2019 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import multiprocessing
import shutil
import warnings
from typing import Any, Dict, List, Optional, Union

import grpc
import pandas as pd

from feast.config import Config
from feast.constants import ConfigOptions as opt
from feast.data_format import ParquetFormat
from feast.data_source import BigQuerySource, FileSource
from feast.entity import Entity
from feast.feature import Feature, FeatureRef, _build_feature_references
from feast.feature_table import FeatureTable
from feast.grpc import auth as feast_auth
from feast.grpc.grpc import create_grpc_channel
from feast.loaders.ingest import (
    _check_field_mappings,
    _read_table_from_source,
    _upload_to_bq_source,
    _upload_to_file_source,
    _write_non_partitioned_table_from_source,
    _write_partitioned_table_from_source,
)
from feast.online_response import OnlineResponse, _infer_online_entity_rows
from feast.protos.feast.core.CoreService_pb2 import (
    ApplyEntityRequest,
    ApplyEntityResponse,
    ApplyFeatureTableRequest,
    ApplyFeatureTableResponse,
    ArchiveProjectRequest,
    ArchiveProjectResponse,
    CreateProjectRequest,
    CreateProjectResponse,
    DeleteFeatureTableRequest,
    GetEntityRequest,
    GetEntityResponse,
    GetFeastCoreVersionRequest,
    GetFeatureTableRequest,
    GetFeatureTableResponse,
    ListEntitiesRequest,
    ListEntitiesResponse,
    ListFeaturesRequest,
    ListFeaturesResponse,
    ListFeatureTablesRequest,
    ListFeatureTablesResponse,
    ListProjectsRequest,
    ListProjectsResponse,
)
from feast.protos.feast.core.CoreService_pb2_grpc import CoreServiceStub
from feast.protos.feast.serving.ServingService_pb2 import (
    GetFeastServingInfoRequest,
    GetOnlineFeaturesRequestV2,
)
from feast.protos.feast.serving.ServingService_pb2_grpc import ServingServiceStub
from feast.registry import Registry
from feast.telemetry import Telemetry

_logger = logging.getLogger(__name__)

CPU_COUNT: int = multiprocessing.cpu_count()

warnings.simplefilter("once", DeprecationWarning)


class Client:
    """
    Feast Client: Used for creating, managing, and retrieving features.
    """

    def __init__(self, options: Optional[Dict[str, str]] = None, **kwargs):
        """
        The Feast Client should be initialized with at least one service url
        Please see constants.py for configuration options. Commonly used options
        or arguments include:
            core_url: Feast Core URL. Used to manage features
            serving_url: Feast Serving URL. Used to retrieve features
            project: Sets the active project. This field is optional.
            core_secure: Use client-side SSL/TLS for Core gRPC API
            serving_secure: Use client-side SSL/TLS for Serving gRPC API
            enable_auth: Enable authentication and authorization
            auth_provider: Authentication provider – "google" or "oauth"
            if auth_provider is "oauth", the following fields are mandatory –
            oauth_grant_type, oauth_client_id, oauth_client_secret, oauth_audience, oauth_token_request_url

        Args:
            options: Configuration options to initialize client with
            **kwargs: Additional keyword arguments that will be used as
                configuration options along with "options"
        """

        if options is None:
            options = dict()
        self._config = Config(options={**options, **kwargs})

        self._core_service_stub: Optional[CoreServiceStub] = None
        self._serving_service_stub: Optional[ServingServiceStub] = None
        self._auth_metadata: Optional[grpc.AuthMetadataPlugin] = None
        self._registry_impl: Optional[Registry] = None

        # Configure Auth Metadata Plugin if auth is enabled
        if self._config.getboolean(opt.ENABLE_AUTH):
            self._auth_metadata = feast_auth.get_auth_metadata_plugin(self._config)

        self._tele = Telemetry()

    @property
    def config(self) -> Config:
        return self._config

    @property
    def _core_service(self):
        """
        Creates or returns the gRPC Feast Core Service Stub

        Returns: CoreServiceStub
        """
        if not self._core_service_stub:
            channel = create_grpc_channel(
                url=self._config.get(opt.CORE_URL),
                enable_ssl=self._config.getboolean(opt.CORE_ENABLE_SSL),
                enable_auth=self._config.getboolean(opt.ENABLE_AUTH),
                ssl_server_cert_path=self._config.get(opt.CORE_SERVER_SSL_CERT),
                auth_metadata_plugin=self._auth_metadata,
                timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
            )
            self._core_service_stub = CoreServiceStub(channel)
        return self._core_service_stub

    @property
    def _use_object_store_registry(self) -> bool:
        return self._config.exists(opt.REGISTRY_PATH)

    @property
    def _registry(self):
        if self._registry_impl is None:
            self._registry_impl = Registry(self._config.get(opt.REGISTRY_PATH))
        return self._registry_impl

    @property
    def _serving_service(self):
        """
        Creates or returns the gRPC Feast Serving Service Stub. If both `opentracing`
        and `grpcio-opentracing` are installed, an opentracing interceptor will be
        instantiated based on the global tracer.

        Returns: ServingServiceStub
        """
        if not self._serving_service_stub:
            channel = create_grpc_channel(
                url=self._config.get(opt.SERVING_URL),
                enable_ssl=self._config.getboolean(opt.SERVING_ENABLE_SSL),
                enable_auth=self._config.getboolean(opt.ENABLE_AUTH),
                ssl_server_cert_path=self._config.get(opt.SERVING_SERVER_SSL_CERT),
                auth_metadata_plugin=self._auth_metadata,
                timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
            )
            try:
                import opentracing
                from grpc_opentracing import open_tracing_client_interceptor
                from grpc_opentracing.grpcext import intercept_channel

                interceptor = open_tracing_client_interceptor(
                    opentracing.global_tracer()
                )
                channel = intercept_channel(channel, interceptor)
            except ImportError:
                pass
            self._serving_service_stub = ServingServiceStub(channel)
        return self._serving_service_stub

    def _extra_grpc_params(self) -> Dict[str, Any]:
        return dict(
            timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
            metadata=self._get_grpc_metadata(),
        )

    @property
    def core_url(self) -> str:
        """
        Retrieve Feast Core URL

        Returns:
            Feast Core URL string
        """
        return self._config.get(opt.CORE_URL)

    @core_url.setter
    def core_url(self, value: str):
        """
        Set the Feast Core URL

        Args:
            value: Feast Core URL
        """
        self._config.set(opt.CORE_URL, value)

    @property
    def serving_url(self) -> str:
        """
        Retrieve Feast Serving URL

        Returns:
            Feast Serving URL string
        """
        return self._config.get(opt.SERVING_URL)

    @serving_url.setter
    def serving_url(self, value: str):
        """
        Set the Feast Serving URL

        Args:
            value: Feast Serving URL
        """
        self._config.set(opt.SERVING_URL, value)

    @property
    def job_service_url(self) -> str:
        """
        Retrieve Feast Job Service URL

        Returns:
            Feast Job Service URL string
        """
        return self._config.get(opt.JOB_SERVICE_URL)

    @job_service_url.setter
    def job_service_url(self, value: str):
        """
        Set the Feast Job Service URL

        Args:
            value: Feast Job Service URL
        """
        self._config.set(opt.JOB_SERVICE_URL, value)

    @property
    def core_secure(self) -> bool:
        """
        Retrieve Feast Core client-side SSL/TLS setting

        Returns:
            Whether client-side SSL/TLS is enabled
        """
        return self._config.getboolean(opt.CORE_ENABLE_SSL)

    @core_secure.setter
    def core_secure(self, value: bool):
        """
        Set the Feast Core client-side SSL/TLS setting

        Args:
            value: True to enable client-side SSL/TLS
        """
        self._config.set(opt.CORE_ENABLE_SSL, value)

    @property
    def serving_secure(self) -> bool:
        """
        Retrieve Feast Serving client-side SSL/TLS setting

        Returns:
            Whether client-side SSL/TLS is enabled
        """
        return self._config.getboolean(opt.SERVING_ENABLE_SSL)

    @serving_secure.setter
    def serving_secure(self, value: bool):
        """
        Set the Feast Serving client-side SSL/TLS setting

        Args:
            value: True to enable client-side SSL/TLS
        """
        self._config.set(opt.SERVING_ENABLE_SSL, value)

    @property
    def job_service_secure(self) -> bool:
        """
        Retrieve Feast Job Service client-side SSL/TLS setting

        Returns:
            Whether client-side SSL/TLS is enabled
        """
        return self._config.getboolean(opt.JOB_SERVICE_ENABLE_SSL)

    @job_service_secure.setter
    def job_service_secure(self, value: bool):
        """
        Set the Feast Job Service client-side SSL/TLS setting

        Args:
            value: True to enable client-side SSL/TLS
        """
        self._config.set(opt.JOB_SERVICE_ENABLE_SSL, value)

    def version(self, sdk_only=False):
        """
        Returns version information from Feast Core and Feast Serving
        """
        import pkg_resources

        try:
            sdk_version = pkg_resources.get_distribution("feast").version
        except pkg_resources.DistributionNotFound:
            sdk_version = "local build"
        if sdk_only:
            return sdk_version

        result = {
            "sdk": {"version": sdk_version},
            "serving": "not configured",
            "core": "not configured",
        }

        if self.serving_url:
            serving_version = self._serving_service.GetFeastServingInfo(
                GetFeastServingInfoRequest(),
                timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
                metadata=self._get_grpc_metadata(),
            ).version
            result["serving"] = {"url": self.serving_url, "version": serving_version}

        if not self._use_object_store_registry and self.core_url:
            core_version = self._core_service.GetFeastCoreVersion(
                GetFeastCoreVersionRequest(),
                timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
                metadata=self._get_grpc_metadata(),
            ).version
            result["core"] = {"url": self.core_url, "version": core_version}

        return result

    @property
    def project(self) -> str:
        """
        Retrieve currently active project

        Returns:
            Project name
        """
        if not self._config.get(opt.PROJECT):
            raise ValueError("No project has been configured.")
        return self._config.get(opt.PROJECT)

    def set_project(self, project: Optional[str] = None):
        """
        Set currently active Feast project

        Args:
            project: Project to set as active. If unset, will reset to the default project.
        """
        if project is None:
            project = opt().PROJECT
        self._config.set(opt.PROJECT, project)

    def list_projects(self) -> List[str]:
        """
        List all active Feast projects

        Returns:
            List of project names

        """

        if self._use_object_store_registry:
            raise NotImplementedError(
                "Projects are not implemented for object store registry."
            )
        else:
            response = self._core_service.ListProjects(
                ListProjectsRequest(),
                timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
                metadata=self._get_grpc_metadata(),
            )  # type: ListProjectsResponse
            return list(response.projects)

    def create_project(self, project: str):
        """
        Creates a Feast project

        Args:
            project: Name of project
        """

        if self._use_object_store_registry:
            raise NotImplementedError(
                "Projects are not implemented for object store registry."
            )
        else:
            self._core_service.CreateProject(
                CreateProjectRequest(name=project),
                timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
                metadata=self._get_grpc_metadata(),
            )  # type: CreateProjectResponse

    def archive_project(self, project):
        """
        Archives a project. Project will still continue to function for
        ingestion and retrieval, but will be in a read-only state. It will
        also not be visible from the Core API for management purposes.

        Args:
            project: Name of project to archive
        """

        if self._use_object_store_registry:
            raise NotImplementedError(
                "Projects are not implemented for object store registry."
            )
        else:
            try:
                self._core_service.ArchiveProject(
                    ArchiveProjectRequest(name=project),
                    timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
                    metadata=self._get_grpc_metadata(),
                )  # type: ArchiveProjectResponse
            except grpc.RpcError as e:
                raise grpc.RpcError(e.details())

            # revert to the default project
            if self._project == project:
                self._project = opt().PROJECT

    def apply(
        self,
        objects: Union[List[Union[Entity, FeatureTable]], Entity, FeatureTable],
        project: str = None,
    ):
        """
        Idempotently registers entities and feature tables with Feast Core. Either a single
        entity or feature table or a list can be provided.

        Args:
            objects: List of entities and/or feature tables that will be registered

        Examples:
            >>> from feast import Client
            >>> from feast.entity import Entity
            >>> from feast.value_type import ValueType
            >>>
            >>> feast_client = Client(core_url="localhost:6565")
            >>> entity = Entity(
            >>>     name="driver_entity",
            >>>     description="Driver entity for car rides",
            >>>     value_type=ValueType.STRING,
            >>>     labels={
            >>>         "key": "val"
            >>>     }
            >>> )
            >>> feast_client.apply(entity)
        """

        self._tele.log("apply")
        if project is None:
            project = self.project

        if not isinstance(objects, list):
            objects = [objects]
        for obj in objects:
            if isinstance(obj, Entity):
                self._apply_entity(project, obj)  # type: ignore
            elif isinstance(obj, FeatureTable):
                self._apply_feature_table(project, obj)  # type: ignore
            else:
                raise ValueError(
                    f"Could not determine object type to apply {obj} with type {type(obj)}. Type must be Entity or FeatureTable."
                )

    def apply_entity(self, entities: Union[List[Entity], Entity], project: str = None):
        """
        Deprecated. Please see apply().
        """
        warnings.warn(
            "The method apply_entity() is being deprecated. Please use apply() instead. Feast 0.10 and onwards will not support apply_entity().",
            DeprecationWarning,
        )

        if project is None:
            project = self.project

        if not isinstance(entities, list):
            entities = [entities]
        for entity in entities:
            if isinstance(entity, Entity):
                self._apply_entity(project, entity)  # type: ignore
                continue
            raise ValueError(f"Could not determine entity type to apply {entity}")

    def _apply_entity(self, project: str, entity: Entity):
        """
        Registers a single entity with Feast

        Args:
            entity: Entity that will be registered
        """

        if self._use_object_store_registry:
            return self._registry.apply_entity(entity, project)
        else:
            entity.is_valid()
            entity_proto = entity.to_spec_proto()

            # Convert the entity to a request and send to Feast Core
            try:
                apply_entity_response = self._core_service.ApplyEntity(
                    ApplyEntityRequest(project=project, spec=entity_proto),  # type: ignore
                    timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
                    metadata=self._get_grpc_metadata(),
                )  # type: ApplyEntityResponse
            except grpc.RpcError as e:
                raise grpc.RpcError(e.details())

            # Extract the returned entity
            applied_entity = Entity.from_proto(apply_entity_response.entity)

            # Deep copy from the returned entity to the local entity
            entity._update_from_entity(applied_entity)

    def list_entities(
        self, project: str = None, labels: Dict[str, str] = dict()
    ) -> List[Entity]:
        """
        Retrieve a list of entities from Feast Core

        Args:
            project: Filter entities based on project name
            labels: User-defined labels that these entities are associated with

        Returns:
            List of entities
        """

        if project is None:
            project = self.project

        if self._use_object_store_registry:
            return self._registry.list_entities(project)
        else:
            filter = ListEntitiesRequest.Filter(project=project, labels=labels)

            # Get latest entities from Feast Core
            entity_protos = self._core_service.ListEntities(
                ListEntitiesRequest(filter=filter), metadata=self._get_grpc_metadata(),
            )  # type: ListEntitiesResponse

            # Extract entities and return
            entities = []
            for entity_proto in entity_protos.entities:
                entity = Entity.from_proto(entity_proto)
                entity._client = self
                entities.append(entity)
            return entities

    def get_entity(self, name: str, project: str = None) -> Entity:
        """
        Retrieves an entity.

        Args:
            project: Feast project that this entity belongs to
            name: Name of entity

        Returns:
            Returns either the specified entity, or raises an exception if
            none is found
        """

        self._tele.log("get_entity")

        if project is None:
            project = self.project

        if self._use_object_store_registry:
            return self._registry.get_entity(name, project)
        else:
            try:
                get_entity_response = self._core_service.GetEntity(
                    GetEntityRequest(project=project, name=name.strip()),
                    metadata=self._get_grpc_metadata(),
                )  # type: GetEntityResponse
            except grpc.RpcError as e:
                raise grpc.RpcError(e.details())
            entity = Entity.from_proto(get_entity_response.entity)

            return entity

    def apply_feature_table(
        self,
        feature_tables: Union[List[FeatureTable], FeatureTable],
        project: str = None,
    ):
        """
        Deprecated. Please see apply().
        """
        warnings.warn(
            "The method apply_feature_table() is being deprecated. Please use apply() instead. Feast 0.10 and onwards will not support apply_feature_table().",
            DeprecationWarning,
        )

        if project is None:
            project = self.project

        if not isinstance(feature_tables, list):
            feature_tables = [feature_tables]
        for feature_table in feature_tables:
            if isinstance(feature_table, FeatureTable):
                self._apply_feature_table(project, feature_table)  # type: ignore
                continue
            raise ValueError(
                f"Could not determine feature table type to apply {feature_table}"
            )

    def _apply_feature_table(self, project: str, feature_table: FeatureTable):
        """
        Registers a single feature table with Feast

        Args:
            feature_table: Feature table that will be registered
        """

        if self._use_object_store_registry:
            return self._registry.apply_feature_table(feature_table, project)
        else:
            feature_table.is_valid()
            feature_table_proto = feature_table.to_spec_proto()

            # Convert the feature table to a request and send to Feast Core
            try:
                apply_feature_table_response = self._core_service.ApplyFeatureTable(
                    ApplyFeatureTableRequest(project=project, table_spec=feature_table_proto),  # type: ignore
                    timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
                    metadata=self._get_grpc_metadata(),
                )  # type: ApplyFeatureTableResponse
            except grpc.RpcError as e:
                raise grpc.RpcError(e.details())

            # Extract the returned feature table
            applied_feature_table = FeatureTable.from_proto(
                apply_feature_table_response.table
            )

            # Deep copy from the returned feature table to the local entity
            feature_table._update_from_feature_table(applied_feature_table)

    def list_feature_tables(
        self, project: str = None, labels: Dict[str, str] = dict()
    ) -> List[FeatureTable]:
        """
        Retrieve a list of feature tables from Feast Core

        Args:
            project: Filter feature tables based on project name

        Returns:
            List of feature tables
        """

        if project is None:
            project = self.project

        if self._use_object_store_registry:
            return self._registry.list_feature_tables(project)
        else:
            filter = ListFeatureTablesRequest.Filter(project=project, labels=labels)

            # Get latest feature tables from Feast Core
            feature_table_protos = self._core_service.ListFeatureTables(
                ListFeatureTablesRequest(filter=filter),
                metadata=self._get_grpc_metadata(),
            )  # type: ListFeatureTablesResponse

            # Extract feature tables and return
            feature_tables = []
            for feature_table_proto in feature_table_protos.tables:
                feature_table = FeatureTable.from_proto(feature_table_proto)
                feature_table._client = self
                feature_tables.append(feature_table)
            return feature_tables

    def get_feature_table(self, name: str, project: str = None) -> FeatureTable:
        """
        Retrieves a feature table.

        Args:
            project: Feast project that this feature table belongs to
            name: Name of feature table

        Returns:
            Returns either the specified feature table, or raises an exception if
            none is found
        """

        self._tele.log("get_feature_table")

        if project is None:
            project = self.project

        if self._use_object_store_registry:
            return self._registry.get_feature_table(name, project)
        else:
            try:
                get_feature_table_response = self._core_service.GetFeatureTable(
                    GetFeatureTableRequest(project=project, name=name.strip()),
                    metadata=self._get_grpc_metadata(),
                )  # type: GetFeatureTableResponse
            except grpc.RpcError as e:
                raise grpc.RpcError(e.details())
            return FeatureTable.from_proto(get_feature_table_response.table)

    def delete_feature_table(self, name: str, project: str = None) -> None:
        """
        Deletes a feature table.

        Args:
            project: Feast project that this feature table belongs to
            name: Name of feature table
        """

        if project is None:
            project = self.project

        if self._use_object_store_registry:
            return self._registry.delete_feature_table(name, project)
        else:
            try:
                self._core_service.DeleteFeatureTable(
                    DeleteFeatureTableRequest(project=project, name=name.strip()),
                    metadata=self._get_grpc_metadata(),
                )
            except grpc.RpcError as e:
                raise grpc.RpcError(e.details())

    def list_features_by_ref(
        self,
        project: str = None,
        entities: List[str] = list(),
        labels: Dict[str, str] = dict(),
    ) -> Dict[FeatureRef, Feature]:
        """
        Retrieve a dictionary of feature reference to feature from Feast Core based on filters provided.

        Args:
            project: Feast project that these features belongs to
            entities: Feast entity that these features are associated with
            labels: Feast labels that these features are associated with

        Returns:
            Dictionary of <feature references: features>

        Examples:
            >>> from feast import Client
            >>>
            >>> feast_client = Client(core_url="localhost:6565")
            >>> features = feast_client.list_features(project="test_project", entities=["driver_id"], labels={"key1":"val1","key2":"val2"})
            >>> print(features)
        """

        if self._use_object_store_registry:
            raise NotImplementedError(
                "This function is not implemented for object store registry."
            )
        else:
            if project is None:
                project = self.project

            filter = ListFeaturesRequest.Filter(
                project=project, entities=entities, labels=labels
            )

            feature_protos = self._core_service.ListFeatures(
                ListFeaturesRequest(filter=filter), metadata=self._get_grpc_metadata(),
            )  # type: ListFeaturesResponse

            # Extract features and return
            features_dict = {}
            for ref_str, feature_proto in feature_protos.features.items():
                feature_ref = FeatureRef.from_str(ref_str)
                feature = Feature.from_proto(feature_proto)
                features_dict[feature_ref] = feature

            return features_dict

    def ingest(
        self,
        feature_table: Union[str, FeatureTable],
        source: Union[pd.DataFrame, str],
        project: str = None,
        chunk_size: int = 10000,
        max_workers: int = max(CPU_COUNT - 1, 1),
        timeout: int = int(opt().BATCH_INGESTION_PRODUCTION_TIMEOUT),
    ) -> None:
        """
        Batch load feature data into a FeatureTable.

        Args:
            feature_table (typing.Union[str, feast.feature_table.FeatureTable]):
                FeatureTable object or the string name of the feature table

            source (typing.Union[pd.DataFrame, str]):
                Either a file path or Pandas Dataframe to ingest into Feast
                Files that are currently supported:
                    * parquet
                    * csv
                    * json

            project: Feast project to locate FeatureTable

            chunk_size (int):
                Amount of rows to load and ingest at a time.

            max_workers (int):
                Number of worker processes to use to encode values.

            timeout (int):
                Timeout in seconds to wait for completion.

        Examples:
            >>> from feast import Client
            >>>
            >>> client = Client(core_url="localhost:6565")
            >>> ft_df = pd.DataFrame(
            >>>         {
            >>>            "datetime": [pd.datetime.now()],
            >>>            "driver": [1001],
            >>>            "rating": [4.3],
            >>>         }
            >>>     )
            >>> client.set_project("project1")
            >>>
            >>> driver_ft = client.get_feature_table("driver")
            >>> client.ingest(driver_ft, ft_df)
        """

        self._tele.log("ingest")
        if project is None:
            project = self.project
        if isinstance(feature_table, str):
            name = feature_table
        if isinstance(feature_table, FeatureTable):
            name = feature_table.name

        fetched_feature_table: Optional[FeatureTable] = self.get_feature_table(
            name, project
        )
        if fetched_feature_table is not None:
            feature_table = fetched_feature_table
        else:
            raise Exception(f"FeatureTable, {name} cannot be found.")

        # Check 1) Only parquet file format for FeatureTable batch source is supported
        if (
            feature_table.batch_source
            and issubclass(type(feature_table.batch_source), FileSource)
            and isinstance(
                type(feature_table.batch_source.file_options.file_format), ParquetFormat
            )
        ):
            raise Exception(
                f"No suitable batch source found for FeatureTable, {name}."
                f"Only BATCH_FILE source with parquet format is supported for batch ingestion."
            )

        pyarrow_table, column_names = _read_table_from_source(source)
        # Check 2) Check if FeatureTable batch source field mappings can be found in provided source table
        _check_field_mappings(
            column_names,
            name,
            feature_table.batch_source.event_timestamp_column,
            feature_table.batch_source.field_mapping,
        )

        dir_path = None
        with_partitions = False
        if (
            issubclass(type(feature_table.batch_source), FileSource)
            and feature_table.batch_source.date_partition_column
        ):
            with_partitions = True
            dest_path = _write_partitioned_table_from_source(
                column_names,
                pyarrow_table,
                feature_table.batch_source.date_partition_column,
                feature_table.batch_source.event_timestamp_column,
            )
        else:
            dir_path, dest_path = _write_non_partitioned_table_from_source(
                column_names, pyarrow_table, chunk_size, max_workers,
            )

        try:
            if issubclass(type(feature_table.batch_source), FileSource):
                file_url = feature_table.batch_source.file_options.file_url.rstrip("*")
                _upload_to_file_source(
                    file_url, with_partitions, dest_path, self._config
                )
            if issubclass(type(feature_table.batch_source), BigQuerySource):
                bq_table_ref = feature_table.batch_source.bigquery_options.table_ref
                feature_table_timestamp_column = (
                    feature_table.batch_source.event_timestamp_column
                )

                _upload_to_bq_source(
                    bq_table_ref, feature_table_timestamp_column, dest_path
                )
        finally:
            # Remove parquet file(s) that were created earlier
            print("Removing temporary file(s)...")
            if dir_path:
                shutil.rmtree(dir_path)

        print("Data has been successfully ingested into FeatureTable batch source.")

    def _get_grpc_metadata(self):
        """
        Returns a metadata tuple to attach to gRPC requests. This is primarily
        used when authentication is enabled but SSL/TLS is disabled.

        Returns: Tuple of metadata to attach to each gRPC call
        """
        if self._config.getboolean(opt.ENABLE_AUTH) and self._auth_metadata:
            return self._auth_metadata.get_signed_meta()
        return ()

    def get_online_features(
        self,
        feature_refs: List[str],
        entity_rows: List[Dict[str, Any]],
        project: Optional[str] = None,
    ) -> OnlineResponse:
        """
        Retrieves the latest online feature data from Feast Serving.
        Args:
            feature_refs: List of feature references that will be returned for each entity.
                Each feature reference should have the following format:
                "feature_table:feature" where "feature_table" & "feature" refer to
                the feature and feature table names respectively.
                Only the feature name is required.
            entity_rows: A list of dictionaries where each key-value is an entity-name, entity-value pair.
            project: Optionally specify the the project override. If specified, uses given project for retrieval.
                Overrides the projects specified in Feature References if also are specified.
        Returns:
            GetOnlineFeaturesResponse containing the feature data in records.
            Each EntityRow provided will yield one record, which contains
            data fields with data value and field status metadata (if included).
        Examples:
            >>> from feast import Client
            >>>
            >>> feast_client = Client(core_url="localhost:6565", serving_url="localhost:6566")
            >>> feature_refs = ["sales:daily_transactions"]
            >>> entity_rows = [{"customer_id": 0},{"customer_id": 1}]
            >>>
            >>> online_response = feast_client.get_online_features(
            >>>     feature_refs, entity_rows, project="my_project")
            >>> online_response_dict = online_response.to_dict()
            >>> print(online_response_dict)
            {'sales:daily_transactions': [1.1,1.2], 'sales:customer_id': [0,1]}
        """

        self._tele.log("get_online_features")
        try:
            response = self._serving_service.GetOnlineFeaturesV2(
                GetOnlineFeaturesRequestV2(
                    features=_build_feature_references(feature_ref_strs=feature_refs),
                    entity_rows=_infer_online_entity_rows(entity_rows),
                    project=project if project is not None else self.project,
                ),
                timeout=self._config.getint(opt.GRPC_CONNECTION_TIMEOUT),
                metadata=self._get_grpc_metadata(),
            )
        except grpc.RpcError as e:
            raise grpc.RpcError(e.details())

        response = OnlineResponse(response)
        return response
